# -*- coding: utf-8 -*-
# ==============================================================================
# Copyright 2018 The TensorFlow Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

import hashlib
import os
import shutil
import sys
import tarfile
import zipfile
from urllib.error import HTTPError, URLError
from urllib.request import urlretrieve

try:
    from tqdm import tqdm as Progbar
except ImportError:

    class Progbar:
        def __init__(self, total_size):
            self.total_size = total_size

        def update(self, size):
            print(".", end="")
            sys.stdout.flush()

        def close(self):
            print()


def round_to_nearest(x, base):
    return base * round(x / base)


def get_file(
    fname,
    origin,
    md5_hash=None,
    file_hash=None,
    cache_subdir="datasets",
    hash_algorithm="auto",
    extract=False,
    archive_format="auto",
    cache_dir=None,
):
    """Downloads a file from a URL if it not already in the cache.

    By default the file at the url `origin` is downloaded to the
    cache_dir `~/.aitk/`, placed in the cache_subdir `datasets`,
    and given the filename `fname`. The final location of a file
    `example.txt` would therefore be `~/.aitk/datasets/example.txt`.

    Files in tar, tar.gz, tar.bz, and zip formats can also be extracted.
    Passing a hash will verify the file after download. The command line
    programs `shasum` and `sha256sum` can compute the hash.

    Arguments:
        fname: Name of the file. If an absolute path `/path/to/file.txt` is
            specified the file will be saved at that location.
        origin: Original URL of the file.
        md5_hash: Deprecated in favor of 'file_hash'.
            md5 hash of the file for verification
        file_hash: The expected hash string of the file after download.
            The sha256 and md5 hash algorithms are both supported.
        cache_subdir: Subdirectory under the aitk cache dir where the file is
            saved. If an absolute path `/path/to/folder` is
            specified the file will be saved at that location.
        hash_algorithm: Select the hash algorithm to verify the file.
            options are 'md5', 'sha256', and 'auto'.
            The default 'auto' detects the hash algorithm in use.
        extract: True tries extracting the file as an Archive, like tar or zip.
        archive_format: Archive format to try for extracting the file.
            Options are 'auto', 'tar', 'zip', and None.
            'tar' includes tar, tar.gz, and tar.bz files.
            The default 'auto' is ['tar', 'zip'].
            None or an empty list will return no matches found.
        cache_dir: Location to store cached files, when None it
            defaults to the aitk default.

    Returns:
        Path to the downloaded file
    """
    if cache_dir is None:
        cache_dir = os.path.join(os.path.expanduser("~"), ".aitk")
    if md5_hash is not None and file_hash is None:
        file_hash = md5_hash
        hash_algorithm = "md5"
    datadir_base = os.path.expanduser(cache_dir)
    if not os.access(datadir_base, os.W_OK):
        datadir_base = os.path.join("/tmp", ".aitk")
    datadir = os.path.join(datadir_base, cache_subdir)
    _makedirs_exist_ok(datadir)

    fpath = os.path.join(datadir, fname)

    download = False
    if os.path.exists(fpath):
        # File found; verify integrity if a hash was provided.
        if file_hash is not None:
            if not validate_file(fpath, file_hash, algorithm=hash_algorithm):
                print(
                    "A local file was found, but it seems to be "
                    "incomplete or outdated because the "
                    + hash_algorithm
                    + " file hash does not match the original value of "
                    + file_hash
                    + " so we will re-download the data."
                )
                download = True
    else:
        download = True

    if download:
        print("Downloading data from", origin)

        class ProgressTracker(object):
            # Maintain progbar for the lifetime of download.
            # This design was chosen for Python 2.7 compatibility.
            progbar = None

        def dl_progress(count, block_size, total_size):
            if ProgressTracker.progbar is None:
                if total_size == -1:
                    total_size = None
                ProgressTracker.progbar = Progbar(total_size)
            else:
                ProgressTracker.progbar.update(count * block_size)

        error_msg = "URL fetch failure on {}: {} -- {}"
        try:
            try:
                urlretrieve(origin, fpath, dl_progress)
            except HTTPError as e:
                raise Exception(error_msg.format(origin, e.code, e.msg))
            except URLError as e:
                raise Exception(error_msg.format(origin, e.errno, e.reason))
        except (Exception, KeyboardInterrupt):
            if os.path.exists(fpath):
                os.remove(fpath)
            raise
        ProgressTracker.progbar.close()
        ProgressTracker.progbar = None

        if extract:
            _extract_archive(fpath, datadir, archive_format)

    return fpath


def _makedirs_exist_ok(datadir):
    os.makedirs(datadir, exist_ok=True)  # pylint: disable=unexpected-keyword-arg


def _extract_archive(file_path, path=".", archive_format="auto"):
    """Extracts an archive if it matches tar, tar.gz, tar.bz, or zip formats.

    Arguments:
        file_path: path to the archive file
        path: path to extract the archive file
        archive_format: Archive format to try for extracting the file.
            Options are 'auto', 'tar', 'zip', and None.
            'tar' includes tar, tar.gz, and tar.bz files.
            The default 'auto' is ['tar', 'zip'].
            None or an empty list will return no matches found.

    Returns:
        True if a match was found and an archive extraction was completed,
        False otherwise.
    """
    if archive_format is None:
        return False
    if archive_format == "auto":
        archive_format = ["tar", "zip"]
    if isinstance(archive_format, (str,)):
        archive_format = [archive_format]

    for archive_type in archive_format:
        if archive_type == "tar":
            open_fn = tarfile.open
            is_match_fn = tarfile.is_tarfile
        if archive_type == "zip":
            open_fn = zipfile.ZipFile
            is_match_fn = zipfile.is_zipfile

        if is_match_fn(file_path):
            with open_fn(file_path) as archive:
                try:
                    archive.extractall(path)
                except (tarfile.TarError, RuntimeError, KeyboardInterrupt):
                    if os.path.exists(path):
                        if os.path.isfile(path):
                            os.remove(path)
                        else:
                            shutil.rmtree(path)
                    raise
            return True
    return False


def _hash_file(fpath, algorithm="sha256", chunk_size=65535):
    """Calculates a file sha256 or md5 hash.

    Example:

    ```python
    _hash_file('/path/to/file.zip')
    'e3b0c44298fc1c149afbf4c8996fb92427ae41e4649b934ca495991b7852b855'
    ```

    Arguments:
        fpath: path to the file being validated
        algorithm: hash algorithm, one of 'auto', 'sha256', or 'md5'.
            The default 'auto' detects the hash algorithm in use.
        chunk_size: Bytes to read at a time, important for large files.

    Returns:
        The file hash
    """
    if (algorithm == "sha256") or (algorithm == "auto" and len(hash) == 64):
        hasher = hashlib.sha256()
    else:
        hasher = hashlib.md5()

    with open(fpath, "rb") as fpath_file:
        for chunk in iter(lambda: fpath_file.read(chunk_size), b""):
            hasher.update(chunk)

    return hasher.hexdigest()


def validate_file(fpath, file_hash, algorithm="auto", chunk_size=65535):
    """Validates a file against a sha256 or md5 hash.

    Arguments:
        fpath: path to the file being validated
        file_hash:  The expected hash string of the file.
            The sha256 and md5 hash algorithms are both supported.
        algorithm: Hash algorithm, one of 'auto', 'sha256', or 'md5'.
            The default 'auto' detects the hash algorithm in use.
        chunk_size: Bytes to read at a time, important for large files.

    Returns:
        Whether the file is valid
    """
    if (algorithm == "sha256") or (algorithm == "auto" and len(file_hash) == 64):
        hasher = "sha256"
    else:
        hasher = "md5"

    if str(_hash_file(fpath, hasher, chunk_size)) == str(file_hash):
        return True
    else:
        return False

class Dataset():
    def __init__(
            self,
            train_inputs=None,
            train_targets=None,
            train_features=None,
            test_inputs=None,
            test_targets=None,
            test_features=None,
    ):
        self.train_inputs = train_inputs or []
        self.train_targets = train_targets or []
        self.train_features = train_features or []
        self.test_inputs = test_inputs or []
        self.test_targets = test_targets or []
        self.test_features = test_features or []

    def select(self, array, indices):
        for i in range(len(array)):
            if i in indices:
                yield array[i]

    def query_train(self, compare_in="or", includes=None, compare_ex="or", excludes=None):
        """
        Select items from the train set.
        
        Args:
            contains: a list of features that a row must contain (OR)
            not_contains: a list of features that a row must not contain (OR)

        Returns a Dataset with selected rows.
        """
        if includes is None:
            includes = []
        if excludes is None:
            excludes = []

        if not isinstance(includes, list) or not isinstance(excludes, list):
            raise Exception("includes and excludes must be lists")

        indices = []
            
        for i in range(len(self.train_inputs)):
            features = self.train_features[i]
            if compare_in == "or":
                include = any([f in features for f in includes])
            elif compare_in == "and":
                include = all([f in features for f in includes])

            if include:
                if compare_ex == "or":
                    include = not any([f in features for f in excludes])
                elif compare_ex == "and":
                    include = not all([f in features for f in excludes])

                if include:
                    indices.append(i)
        
        return Dataset(
            list(self.select(self.train_inputs, indices)),
            list(self.select(self.train_targets, indices)),
            list(self.select(self.train_features, indices)),
        )

    def query_test(self, compare_in="or", includes=None, compare_ex="or", excludes=None):
        """
        Select items from the test set.
        
        Args:
            contains: a list of features that a row must contain
            not_contains: a list of features that a row must not contain

        Returns a Dataset with selected rows.
        """
        if includes is None:
            includes = []
        if excludes is None:
            excludes = []

        if not isinstance(includes, list) or not isinstance(excludes, list):
            raise Exception("includes and excludes must be lists")

        indices = []
            
        for i in range(len(self.test_inputs)):
            features = self.test_features[i]
            if includes is not None:
                if compare_in == "or":
                    include = any([f in features for f in includes])
                elif compare_in == "and":
                    include = all([f in features for f in includes])
                else:
                    raise Exception("compare_in must be `or`, or `and`")
            else:
                include = True

            if include:
                if excludes is not None:
                    if compare_ex == "or":
                        include = not any([f in features for f in excludes])
                    elif compare_ex == "and":
                        include = not all([f in features for f in excludes])
                    else:
                        raise Exception("compare_ex must be `or`, or `and`")

                if include:
                    indices.append(i)
        
        return Dataset(
            test_inputs=list(self.select(self.test_inputs, indices)),
            test_targets=list(self.select(self.test_targets, indices)),
            test_features=list(self.select(self.test_features, indices)),
        )



"""Implements the core evolution algorithm."""
from __future__ import print_function

from neat.math_util import mean
from neat.reporting import ReporterSet

import copy

class CompleteExtinctionException(Exception):
    pass


class Population(object):
    """
    This class implements the core evolution algorithm:
        1. Evaluate fitness of all genomes.
        2. Check to see if the termination criterion is satisfied; exit if it is.
        3. Generate the next generation from the current population.
        4. Partition the new generation into species based on genetic similarity.
        5. Go to 1.
    """

    def __init__(self, config, initial_state=None):
        self.reporters = ReporterSet()
        self.config = config
        stagnation = config.stagnation_type(config.stagnation_config, self.reporters)
        self.reproduction = config.reproduction_type(config.reproduction_config,
                                                     self.reporters,
                                                     stagnation)
        if config.fitness_criterion == 'max':
            self.fitness_criterion = max
        elif config.fitness_criterion == 'min':
            self.fitness_criterion = min
        elif config.fitness_criterion == 'mean':
            self.fitness_criterion = mean
        elif not config.no_fitness_termination:
            raise RuntimeError(
                "Unexpected fitness_criterion: {0!r}".format(config.fitness_criterion))

        if initial_state is None:
            # Create a population from scratch, then partition into species.
            self.population = self.reproduction.create_new(config.genome_type,
                                                           config.genome_config,
                                                           config.pop_size)
            self.species = config.species_set_type(config.species_set_config, self.reporters)
            self.generation = 0
            self.species.speciate(config, self.population, self.generation)
            self.best_genome = None
        else:
            self.population, self.species, self.best_genome, self.generation = initial_state


    def add_reporter(self, reporter):
        self.reporters.add(reporter)

    def remove_reporter(self, reporter):
        self.reporters.remove(reporter)

    def run(self, fitness_function, n=None):
        """
        Runs NEAT's genetic algorithm for at most n generations.  If n
        is None, run until solution is found or extinction occurs.

        The user-provided fitness_function must take only two arguments:
            1. The population as a list of (genome id, genome) tuples.
            2. The current configuration object.

        The return value of the fitness function is ignored, but it must assign
        a Python float to the `fitness` member of each genome.

        The fitness function is free to maintain external state, perform
        evaluations in parallel, etc.

        It is assumed that fitness_function does not modify the list of genomes,
        the genomes themselves (apart from updating the fitness member),
        or the configuration object.
        """

        if self.config.no_fitness_termination and (n is None):
            raise RuntimeError("Cannot have no generational limit with no fitness termination")

        k = 0
        while n is None or k < n:
            k += 1

            self.reporters.start_generation(self.generation)

            # Evaluate all genomes using the user-provided function.
            fitness_function(list(self.population.items()), self.config)

            # Gather and report statistics.
            best = None
            for g in self.population.values():
                if g.fitness is None:
                    raise RuntimeError("Fitness not assigned to genome {}".format(g.key))

                if best is None or g.fitness > best.fitness:
                    best = g
            self.reporters.post_evaluate(self.config, self.population, self.species, best)

            # Track the best genome ever seen.
            if self.best_genome is None or best.fitness > self.best_genome.fitness:
                self.best_genome = copy.copy(best)

            if not self.config.no_fitness_termination:
                # End if the fitness threshold is reached.
                fv = self.fitness_criterion(g.fitness for g in self.population.values())
                if fv >= self.config.fitness_threshold:
                    self.reporters.found_solution(self.config, self.generation, best)
                    break

            # Create the next generation from the current generation.
            self.population = self.reproduction.reproduce(self.config, self.species,
                                                          self.config.pop_size, self.generation)

            # Check for complete extinction.
            if not self.species.species:
                self.reporters.complete_extinction()

                # If requested by the user, create a completely new population,
                # otherwise raise an exception.
                if self.config.reset_on_extinction:
                    self.population = self.reproduction.create_new(self.config.genome_type,
                                                                   self.config.genome_config,
                                                                   self.config.pop_size)
                else:
                    raise CompleteExtinctionException()

            # Divide the new population into species.
            self.species.speciate(self.config, self.population, self.generation)

            self.reporters.end_generation(self.config, self.population, self.species)

            self.generation += 1

        if self.config.no_fitness_termination:
            self.reporters.found_solution(self.config, self.generation, self.best_genome)

        return self.best_genome



"""Handles the continuous-time recurrent neural network implementation."""
from __future__ import division

from ..graphs import required_for_output


class CTRNNNodeEval(object):
    def __init__(self, time_constant, activation, aggregation, bias, response, links):
        self.time_constant = time_constant
        self.activation = activation
        self.aggregation = aggregation
        self.bias = bias
        self.response = response
        self.links = links


class CTRNN(object):
    """Sets up the ctrnn network itself."""
    def __init__(self, inputs, outputs, node_evals):
        self.input_nodes = inputs
        self.output_nodes = outputs
        self.node_evals = node_evals

        self.values = [{}, {}]
        for v in self.values:
            for k in inputs + outputs:
                v[k] = 0.0

            for node, ne in self.node_evals.items():
                v[node] = 0.0
                for i, w in ne.links:
                    v[i] = 0.0

        self.active = 0
        self.time_seconds = 0.0

    def reset(self):
        self.values = [dict((k, 0.0) for k in v) for v in self.values]
        self.active = 0
        self.time_seconds = 0.0

    def set_node_value(self, node_key, value):
        for v in self.values:
            v[node_key] = value

    def get_max_time_step(self):  # pragma: no cover
        # TODO: Compute max time step that is known to be numerically stable for
        # the current network configuration.
        # pylint: disable=no-self-use
        raise NotImplementedError()

    def advance(self, inputs, advance_time, time_step=None):
        """
        Advance the simulation by the given amount of time, assuming that inputs are
        constant at the given values during the simulated time.
        """
        final_time_seconds = self.time_seconds + advance_time

        # Use half of the max allowed time step if none is given.
        if time_step is None:  # pragma: no cover
            time_step = 0.5 * self.get_max_time_step()

        if len(self.input_nodes) != len(inputs):
            raise RuntimeError("Expected {0} inputs, got {1}".format(len(self.input_nodes), len(inputs)))

        while self.time_seconds < final_time_seconds:
            dt = min(time_step, final_time_seconds - self.time_seconds)

            ivalues = self.values[self.active]
            ovalues = self.values[1 - self.active]
            self.active = 1 - self.active

            for i, v in zip(self.input_nodes, inputs):
                ivalues[i] = v
                ovalues[i] = v

            for node_key, ne in self.node_evals.items():
                node_inputs = [ivalues[i] * w for i, w in ne.links]
                s = ne.aggregation(node_inputs)
                z = ne.activation(ne.bias + ne.response * s)
                ovalues[node_key] += dt / ne.time_constant * (-ovalues[node_key] + z)

            self.time_seconds += dt

        ovalues = self.values[1 - self.active]
        return [ovalues[i] for i in self.output_nodes]

    @staticmethod
    def create(genome, config, time_constant):
        """ Receives a genome and returns its phenotype (a CTRNN). """
        genome_config = config.genome_config
        required = required_for_output(genome_config.input_keys, genome_config.output_keys, genome.connections)

        # Gather inputs and expressed connections.
        node_inputs = {}
        for cg in genome.connections.values():
            if not cg.enabled:
                continue

            i, o = cg.key
            if o not in required and i not in required:
                continue

            if o not in node_inputs:
                node_inputs[o] = [(i, cg.weight)]
            else:
                node_inputs[o].append((i, cg.weight))

        node_evals = {}
        for node_key, inputs in node_inputs.items():
            node = genome.nodes[node_key]
            activation_function = genome_config.activation_defs.get(node.activation)
            aggregation_function = genome_config.aggregation_function_defs.get(node.aggregation)
            node_evals[node_key] = CTRNNNodeEval(time_constant,
                                                 activation_function,
                                                 aggregation_function,
                                                 node.bias,
                                                 node.response,
                                                 inputs)

        return CTRNN(genome_config.input_keys, genome_config.output_keys, node_evals)



# -*- coding: utf-8 -*-
# ************************************************************
# aitk.robots: Python robot simulator
#
# Copyright (c) 2021 AITK Developers
#
# https://github.com/ArtificialIntelligenceToolkit/aitk.robots
# ************************************************************

import glob
import io
import json
import math
import os
from collections import OrderedDict
from datetime import datetime, timedelta
from functools import wraps
import random

try:
    import numpy as np
except ImportError:
    np = None

from .config import get_aitk_search_paths
from .hit import Hit

PI_OVER_180 = math.pi / 180
PI_OVER_2 = math.pi / 2
ONE80_OVER_PI = 180 / math.pi
TWO_PI = math.pi * 2

MESSAGES = set()

try:
    _in_colab = 'google.colab' in str(get_ipython())
except Exception:
    _in_colab = False

try:
    from IPython.display import display
except ImportError:
    display = print


def in_colab():
    return _in_colab

class Food():
    def __init__(self, x, y, standard_deviation, state):
        self.x = x
        self.y = y
        self.standard_deviation = standard_deviation
        self.state = state

def print_once(message):
    if message not in MESSAGES:
        print(message)
        MESSAGES.add(message)

def compare(a, b):
    if a > b:
        return 1
    elif a < b:
        return -1
    else:
        return 0

def in_range(delta, current, stop_value):
    if delta < 0:
        return current >= stop_value
    elif delta > 0:
        return current <= stop_value
    else:
        return True

def normal_dist(x , mean , sd):
    prob_density = (math.pi * sd) * math.exp(-0.5 * ((x - mean) / sd) ** 2)
    return prob_density

def round_to(value, base):
    return base * round(value / base)

def uniform_angle(angle):
    return angle % TWO_PI

def degrees_to_world(degrees):
    return ((TWO_PI - (degrees * PI_OVER_180)) % TWO_PI)

def world_to_degrees(direction):
    return (((direction + TWO_PI) * -ONE80_OVER_PI) % 360)

def cast_ray(world, robot, x1, y1, a, maxRange,
             x2=None, y2=None, ignore_robots=None):
    # walls and robots
    hits = []
    if x2 is None:
        x2 = math.sin(a) * maxRange + x1
    if y2 is None:
        y2 = math.cos(a) * maxRange + y1

    for wall in world._walls:
        # never detect hit with yourself
        if robot is not None and wall.robot is robot:
            continue
        # ignore this robot:
        if ((ignore_robots is not None) and
            (wall.robot is not None) and
            (wall.robot in ignore_robots)):
            continue
        for line in wall.lines:
            p1 = line.p1
            p2 = line.p2
            pos = intersect_hit(x1, y1, x2, y2, p1.x, p1.y, p2.x, p2.y)
            if pos is not None:
                dist = distance(pos[0], pos[1], x1, y1)
                height = 1.0 if wall.robot is None else wall.robot.height
                color = wall.robot.color if wall.robot else wall.color
                boundary = len(wall.lines) == 1
                hits.append(
                    Hit(wall,
                        wall.robot,
                        height,
                        pos[0],
                        pos[1],
                        dist,
                        color,
                        x1,
                        y1,
                        boundary,
                        a,
                    )
                )

    hits.sort(
        key=lambda a: a.distance, reverse=True
    )  # further away first, back to front
    return hits

def rotate_around(x1, y1, length, angle):
    """
    Swing a line around a point.
    """
    return [x1 + length * math.cos(-angle),
            y1 - length * math.sin(-angle)]

def dot(v, w):
    x, y, z = v
    X, Y, Z = w
    return x * X + y * Y + z * Z


def length(v):
    x, y, z = v
    return math.sqrt(x * x + y * y + z * z)


def vector(b, e):
    x, y, z = b
    X, Y, Z = e
    return (X - x, Y - y, Z - z)


def unit(v):
    x, y, z = v
    mag = length(v)
    return (x / mag, y / mag, z / mag)


def scale(v, sc):
    x, y, z = v
    return (x * sc, y * sc, z * sc)


def add(v, w):
    x, y, z = v
    X, Y, Z = w
    return (x + X, y + Y, z + Z)


def ccw(ax, ay, bx, by, cx, cy):
    # counter clockwise
    return ((cy - ay) * (bx - ax)) > ((by - ay) * (cx - ax))


def intersect(ax, ay, bx, by, cx, cy, dx, dy):
    # Return true if line segments AB and CD intersect
    return ccw(ax, ay, cx, cy, dx, dy) != ccw(bx, by, cx, cy, dx, dy) and (
        ccw(ax, ay, bx, by, cx, cy) != ccw(ax, ay, bx, by, dx, dy)
    )


def coefs(p1x, p1y, p2x, p2y):
    A = p1y - p2y
    B = p2x - p1x
    C = p1x * p2y - p2x * p1y
    return [A, B, -C]


def intersect_coefs(L1_0, L1_1, L1_2, L2_0, L2_1, L2_2):
    D = L1_0 * L2_1 - L1_1 * L2_0
    if D != 0:
        Dx = L1_2 * L2_1 - L1_1 * L2_2
        Dy = L1_0 * L2_2 - L1_2 * L2_0
        x1 = Dx / D
        y1 = Dy / D
        return [x1, y1]
    else:
        return None


def intersect_hit(p1x, p1y, p2x, p2y, p3x, p3y, p4x, p4y):
    """
    Compute the intersection between two lines.
    """
    # http:##stackoverflow.com/questions/20677795/find-the-point-of-intersecting-lines
    L1 = coefs(p1x, p1y, p2x, p2y)
    L2 = coefs(p3x, p3y, p4x, p4y)
    xy = intersect_coefs(L1[0], L1[1], L1[2], L2[0], L2[1], L2[2])
    # now check to see on both segments:
    if xy:
        lowx = min(p1x, p2x) - 0.1
        highx = max(p1x, p2x) + 0.1
        lowy = min(p1y, p2y) - 0.1
        highy = max(p1y, p2y) + 0.1
        if (lowx <= xy[0] and xy[0] <= highx) and (lowy <= xy[1] and xy[1] <= highy):
            lowx = min(p3x, p4x) - 0.1
            highx = max(p3x, p4x) + 0.1
            lowy = min(p3y, p4y) - 0.1
            highy = max(p3y, p4y) + 0.1
            if lowx <= xy[0] and xy[0] <= highx and lowy <= xy[1] and xy[1] <= highy:
                return xy
    return None


def format_time(time):
    hours = time // 3600
    minutes = (time % 3600) // 60
    seconds = (time % 3600) % 60
    return "%02d:%02d:%05.2f" % (hours, minutes, seconds)


def load_world(filename=None):
    """
    worlds/
        test1/
            worlds/test1/w1.json
            worlds/test1/w2.json
        test2/
            worlds/test2/w1.json
        worlds/w1.json

    """
    from .world import World

    if filename is None:
        print("Searching for aitk.robots config files...")
        for path in get_aitk_search_paths():
            print("Directory:", path)
            files = sorted(
                glob.glob(os.path.join(path, "**", "*.json"), recursive=True),
                key=lambda filename: (filename.count("/"), filename),
            )
            if len(files) > 0:
                for fname in files:
                    basename = os.path.splitext(fname)[0]
                    print("    %r" % basename[len(path) :])
            else:
                print("    no files found")
    else:
        if not filename.endswith(".json"):
            filename += ".json"
        for path in get_aitk_search_paths():
            path_filename = os.path.join(path, filename)
            if os.path.exists(path_filename):
                print("Loading %s..." % path_filename)
                with open(path_filename) as fp:
                    contents = fp.read()
                    config = json.loads(contents)
                    config["filename"] = path_filename
                    world = World(**config)
                    return world
        print("No such world found: %r" % filename)
    return None


def load_image(filename, width=None, height=None):
    from PIL import Image

    pathname = find_resource(filename)
    if pathname is not None:
        image = Image.open(pathname)
        if width is not None and height is not None:
            image = image.resize((width, height))
        return image


def find_resource(filename=None):
    if filename is None:
        print("Searching for aitk.robots files...")
        for path in get_aitk_search_paths():
            files = sorted(glob.glob(os.path.join(path, "*.*")))
            print("Directory:", path)
            if len(files) > 0:
                for filename in files:
                    print("    %r" % filename)
            else:
                print("    no files found")
    else:
        for path in get_aitk_search_paths():
            path_filename = os.path.abspath(os.path.join(path, filename))
            if os.path.exists(path_filename):
                return path_filename
        print("No such file found: %r" % filename)
    return None


def image_to_png(image):
    with io.BytesIO() as fp:
        image.save(fp, format="png")
        return fp.getvalue()


def image_to_gif(image):
    # Leave fp opened
    from PIL import Image

    fp = io.BytesIO()
    image.save(fp, "gif")
    frame = Image.open(fp)
    return frame


def gallery(*images, border_width=1, background_color=(255, 255, 255)):
    """
    Construct a gallery of images
    """
    try:
        from PIL import Image
    except ImportError:
        print("gallery() requires Pillow, Python Image Library (PIL)")
        return

    gallery_cols = math.ceil(math.sqrt(len(images)))
    gallery_rows = math.ceil(len(images) / gallery_cols)

    size = images[0].size
    size = size[0] + (border_width * 2), size[1] + (border_width * 2)

    gallery_image = Image.new(
        mode="RGBA",
        size=(int(gallery_cols * size[0]), int(gallery_rows * size[1])),
        color=background_color,
    )

    for i, image in enumerate(images):
        if image.mode != "RGBA":
            image = image.convert("RGBA")
        location = (
            int((i % gallery_cols) * size[0]) + border_width,
            int((i // gallery_cols) * size[1]) + border_width,
        )
        gallery_image.paste(image, location)
    return gallery_image


class arange:
    def __init__(self, start, stop, step):
        self.start = start
        self.stop = stop
        self.step = step

    def __iter__(self):
        current = self.start
        if self.step > 0:
            while current <= self.stop:
                yield current
                current += self.step
        else:
            while current >= self.stop:
                yield current
                current += self.step

    def __len__(self):
        return int(abs(self.stop - self.start) / abs(self.step))


def distance(x1, y1, x2=None, y2=None):
    """
    Takes either (x, y) from two points, or two points
    and returns the distance between them.
    """
    if x2 is None:
        if np is not None:
            # Works with more than two dimensions:
            a1 = np.array(x1)
            a2 = np.array(y1)
            return np.sqrt(np.sum((a1 - a2) ** 2))
        else:
            x1, y1, x2, y2 = x1[0], x1[1], y1[0], y1[1]

    d1 = (x1 - x2)
    d2 = (y1 - y2)
    return math.sqrt(d1 * d1 + d2 * d2)

def distance_point_to_line_3d(point, line_start, line_end):
    """
    Compute distance and location to closest point
    on a line segment in 3D.
    """
    line_vec = vector(line_start, line_end)
    point_vec = vector(line_start, point)
    line_len = length(line_vec)
    line_unitvec = unit(line_vec)
    point_vec_scaled = scale(point_vec, 1.0 / line_len)
    t = dot(line_unitvec, point_vec_scaled)
    if t < 0.0:
        t = 0.0
    elif t > 1.0:
        t = 1.0
    nearest = scale(line_vec, t)
    # Optipization: assume 2D:
    dist = distance(nearest[0], nearest[1], point_vec[0], point_vec[1])
    nearest = add(nearest, line_start)
    return (dist, nearest)


def distance_point_to_line(point, line_start, line_end):
    return distance_point_to_line_3d(
        (point[0], point[1], 0),
        (line_start[0], line_start[1], 0),
        (line_end[0], line_end[1], 0),
    )


def json_dump(config, fp, sort_keys=True, indent=4):
    dumps(fp, config, sort_keys=sort_keys, indent=indent)


def dumps(fp, obj, level=0, sort_keys=True, indent=4, newline="\n", space=" "):
    if isinstance(obj, dict):
        if sort_keys:
            obj = OrderedDict({key: obj[key] for key in sorted(obj.keys())})
        fp.write(newline + (space * indent * level) + "{" + newline)
        comma = ""
        for key, value in obj.items():
            fp.write(comma)
            comma = "," + newline
            fp.write(space * indent * (level + 1))
            fp.write('"%s":%s' % (key, space))
            dumps(fp, value, level + 1, sort_keys, indent, newline, space)
        fp.write(newline + (space * indent * level) + "}")
    elif isinstance(obj, str):
        fp.write('"%s"' % obj)
    elif isinstance(obj, (list, tuple)):
        if len(obj) == 0:
            fp.write("[]")
        else:
            fp.write(newline + (space * indent * level) + "[")
            # fp.write("[")
            comma = ""
            for item in obj:
                fp.write(comma)
                comma = ", "
                dumps(fp, item, level + 1, sort_keys, indent, newline, space)
            # each on their own line
            if len(obj) > 2:
                fp.write(newline + (space * indent * level))
            fp.write("]")
    elif isinstance(obj, bool):
        fp.write("true" if obj else "false")
    elif isinstance(obj, int):
        fp.write(str(obj))
    elif obj is None:
        fp.write("null")
    elif isinstance(obj, float):
        fp.write("%.7g" % obj)
    else:
        raise TypeError("Unknown object %r for json serialization" % obj)


class throttle(object):
    """
    Decorator that prevents a function from being called more than once every
    time period.
    To create a function that cannot be called more than once a minute:
        @throttle(minutes=1)
        def my_fun():
            pass
    """

    def __init__(self, seconds=0, minutes=0, hours=0):
        self.throttle_period = timedelta(seconds=seconds, minutes=minutes, hours=hours)
        self.time_of_last_call = datetime.min

    def __call__(self, fn):
        @wraps(fn)
        def wrapper(*args, **kwargs):
            now = datetime.now()
            time_since_last_call = now - self.time_of_last_call

            if time_since_last_call > self.throttle_period:
                self.time_of_last_call = now
                return fn(*args, **kwargs)

        return wrapper


class Point:
    def __init__(self, x, y):
        self.x = x
        self.y = y

    def __getitem__(self, item):
        if item == 0:
            return self.x
        elif item == 1:
            return self.y

    def __len__(self):
        return 2

    def __repr__(self):
        return "Point(%s,%s)" % (self.x, self.y)

    def copy(self):
        return Point(self.x, self.y)


class Line:
    def __init__(self, p1, p2):
        self.p1 = p1
        self.p2 = p2

    def __repr__(self):
        return "Line(%s,%s)" % (self.p1, self.p2)

class Grid:
    def __init__(self, width, height, step=10):
        self.width = width
        self.height = height
        self.step = step
        self.blocked = {}
        self.grid = self.spread([])
        self.need_update = True

    def clear_walls(self):
        self.blocked = {}
        self.need_update = True

    def block_area(self, x1, y1, x2, y2, box=True):
        if box:
            x_min = round_to(min(x1, x2), self.step)
            y_min = round_to(min(y1, y2), self.step)
            x_max = round_to(max(x1, x2), self.step)
            y_max = round_to(max(y1, y2), self.step)
            for x in range(x_min, x_max):
                for y in range(y_min, y_max):
                    self.blocked[(x,y)] = 1
        else:
            x_start = round_to(x1, self.step)
            x_stop = round_to(x2, self.step)
            y_start = round_to(y1, self.step)
            y_stop = round_to(y2, self.step)

            dx = compare(x_stop, x_start)
            dy = compare(y_stop, y_start)

            max_range = max(abs(x_stop - x_start),
                            abs(y_stop - y_start))

            if dx != 0 and dy != 0:
                x = x_start
                y = y_start
                while in_range(dx, x, x_stop) and in_range(dy, y, y_stop):
                    self.blocked[(round(x),round(y))] = 1
                    self.blocked[(round(x)-1,round(y))] = 1
                    self.blocked[(round(x),round(y)-1)] = 1
                    self.blocked[(round(x)+1,round(y))] = 1
                    self.blocked[(round(x),round(y)+1)] = 1
                    x += dx / max_range * abs(x_stop - x_start)
                    y += dy / max_range * abs(y_stop - y_start)

    def expand(self, visited, start, x, y, dist):
        retval = []
        for dx in [-self.step, 0, self.step]:
            for dy in [-self.step, 0, self.step]:
                if ((0 <= (x + dx) < self.width) and
                    (0 <= (y + dy) < self.height)):
                    if (x + dx, y + dy) not in self.blocked and (x + dx, y + dy) not in visited:
                        # artifacts on ordering:
                        #retval.append((x + dx, y + dy, dist + distance(0, 0, dx, dy)))
                        retval.append((x + dx, y + dy, dist + self.step))
        return retval

    def bfs(self, start):
        queue = [start + (0,)] # distance
        visited = {}
        while len(queue) > 0:
            x, y, dist = queue.pop(0)
            if (x,y) not in visited:
                queue.extend(self.expand(visited, start, x, y, dist))
                visited[(x,y)] = dist
        return visited

    def spread(self, points):
        # points are Food(x, y, standard_deviation, state)
        # Initialize grid with no values:
        grid = [[0 for y in range(self.height)]
                for x in range(self.width)]

        # For each point, search out from there:
        for food in points:
            if food.state == "off":
                continue
            # make sure we search from a spot on the
            # grid
            px = round_to(food.x, self.step)
            py = round_to(food.y, self.step)
            values = self.bfs((px, py))
            for x in range(0, self.width, self.step):
                for y in range(0, self.height, self.step):
                    dist = values.get((x,y), None)
                    if dist is not None:
                        value = self.weight(dist, food.standard_deviation)
                        for i in range(y, y + self.step):
                            for j in range(x, x + self.step):
                                if 0 <= i < self.height and 0 <= j < self.width:
                                    if (j, i) not in self.blocked:
                                        grid[j][i] += value
                                        grid[j][i] = min(grid[j][i], 1.0)
        return grid

    def weight(self, dist, sd=1):
        """
        Return a value between 1 and 0 where
        dist=0 gives 1, and as dist increases,
        the return value falls off in a normal
        distribution.
        """
        return (normal_dist(dist, 0, sd) / math.pi) / sd

    def update(self, blooms):
        """
        Update the grid with blooms of values
        to spread the values over the grid.
        blooms are a list of Food(x, y, standard_deviation, state).
        """
        if self.need_update:
            self.grid = self.spread(blooms)
            self.need_update = False

    def update_walls(self, walls):
        # update the grid to block smells
        self.clear_walls()
        for wall in walls:
            self.update_wall(wall)

    def update_wall(self, wall):
        self.need_update = True
        if wall.robot is None:
            if len(wall.lines) == 4: # box
                p1 = wall.lines[0].p1
                p3 = wall.lines[1].p2
                self.block_area(p1[0], p1[1], p3[0], p3[1], box=True)
            else: # line
                # FIXME: Includes boundary walls!
                for line in wall.lines:
                    self.block_area(
                        line.p1[0],
                        line.p1[1],
                        line.p2[0],
                        line.p2[1],
                        box=False)

    def get(self, x, y):
        """
        Get the reading at the grid point.
        """
        x = max(min(self.width - 1, int(x)), 0)
        y = max(min(self.height - 1, int(y)), 0)
        return self.grid[x][y]

    def get_image(self):
        """
        Get an image of the grid.
        """
        from PIL import Image
        import numpy as np

        data = np.array(self.grid).transpose() * 255
        data = data.astype(np.uint8)
        image = Image.fromarray(data, "L")
        return image



# -*- coding: utf-8 -*-
# ************************************************************
# aitk.robots: Python robot simulator
#
# Copyright (c) 2021 AITK Developers
#
# https://github.com/ArtificialIntelligenceToolkit/aitk.robots
# ************************************************************

class BaseDevice():
    def verify_config(self, valid_keys, config):
        config_keys = set(list(config.keys()))
        extra_keys = config_keys - valid_keys

        if len(extra_keys) > 0:
            raise TypeError("invalid key(s) for device: %r" % extra_keys)



# -*- coding: utf-8 -*-
# *************************************
# aitk.robots: Python robot simulator
#
# Copyright (c) 2020 Calysto Developers
#
# https://github.com/ArtificialIntelligenceToolkit/aitk.robots
#
# *************************************

"""
Classic cart-pole system implemented by Rich Sutton et al.
Copied from http://incompleteideas.net/sutton/book/code/pole.c
permalink: https://perma.cc/C9ZM-652R

Python version based on:
https://raw.githubusercontent.com/openai/gym/master/gym/envs/classic_control/cartpole.py
"""

import math
import random

from .world import World, Robot
from .utils import rotate_around

from aitk.utils import Color

BLUE = Color("blue")
BLACK = Color("black")
YELLOW = Color(207, 166, 54)
GREY = Color("grey")

class Cart(Robot):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.masscart = 1.0
        self.masspole = 0.1
        self.total_mass = (self.masspole + self.masscart)
        self.length = 0.5  # actually half the pole's length
        self.polemass_length = (self.masspole * self.length)
        self.force_mag = 10.0
        self.kinematics_integrator = 'euler'
        self.reset()

    def reset(self):
        # cart_state is: (x, x_dot, theta, theta_dot)
        self.cart_state = [(0.05 - random.random() * 0.1) for i in range(4)]
        self.action = 0
        self._set_pose(0, 0, 0)
        self.reward = 0

    def move(self, action):
        self.action = action

    def _step(self, time_step):
        # action is 0, 1 or -1
        x, x_dot, theta, theta_dot = self.cart_state
        if self.action == 1:
            force = self.force_mag
        elif self.action == -1:
            force = -self.force_mag
        else:
            force = 0
        costheta = math.cos(theta)
        sintheta = math.sin(theta)

        # For the interested reader:
        # https://coneural.org/florian/papers/05_cart_pole.pdf
        temp = (force + self.polemass_length * theta_dot ** 2 * sintheta) / self.total_mass
        thetaacc = (self.world.gravity * sintheta - costheta * temp) / (self.length * (4.0 / 3.0 - self.masspole * costheta ** 2 / self.total_mass))
        xacc = temp - self.polemass_length * thetaacc * costheta / self.total_mass

        if self.kinematics_integrator == 'euler':
            x = x + time_step * x_dot
            x_dot = x_dot + time_step * xacc
            theta = theta + time_step * theta_dot
            theta_dot = theta_dot + time_step * thetaacc
        else:  # semi-implicit euler
            x_dot = x_dot + time_step * xacc
            x = x + time_step * x_dot
            theta_dot = theta_dot + time_step * thetaacc
            theta = theta + time_step * theta_dot

        self.x = x
        self.tvx = x_dot
        self.a = theta
        self.tva = theta_dot
        self.cart_state = (x, x_dot, theta, theta_dot)

    def get_observation(self):
        return self.cart_state

    def update(self, draw_list=None):
        pass

    def draw(self, backend):
        cart_x = self.x * 10
        backend.strokeStyle(GREY, 1)
        backend.draw_line(0, self.world.height/2, self.world.width, self.world.height/2)
        backend.set_fill_style(BLACK)
        backend.draw_rect(self.world.width/2 + cart_x - 25,
                          self.world.height/2 - 10,
                          50,
                          20)
        # draw pole
        backend.strokeStyle(YELLOW, 5)
        x, y = rotate_around(self.world.width/2 + cart_x,
                             self.world.height/2,
                             100, self.cart_state[2] - math.pi/2)
        backend.draw_line(self.world.width/2 + cart_x,
                          self.world.height/2,
                          x, y)
        backend.set_fill_style(BLUE)
        backend.strokeStyle(BLUE, 1)
        backend.draw_circle(self.world.width/2 + cart_x, self.world.height/2, 5)


class CartWorld(World):
    """
    Description:
        A pole is attached by an un-actuated joint to a cart, which moves along
        a frictionless track. The pendulum starts upright, and the goal is to
        prevent it from falling over by increasing and reducing the cart's
        velocity.

    Source:
        This environment corresponds to the version of the cart-pole problem
        described by Barto, Sutton, and Anderson

    Observation:
        Type: Box(4)
        Num     Observation               Min                     Max
        0       Cart Position             -4.8                    4.8
        1       Cart Velocity             -Inf                    Inf
        2       Pole Angle                -0.418 rad (-24 deg)    0.418 rad (24 deg)
        3       Pole Angular Velocity     -Inf                    Inf

    Actions:
        Type: Discrete(2)
        Num   Action
        -1    Push cart to the left
         1    Push cart to the right

        Note: The amount the velocity that is reduced or increased is not
        fixed; it depends on the angle the pole is pointing. This is because
        the center of gravity of the pole increases the amount of energy needed
        to move the cart underneath it

    Starting State:
        All observations are assigned a uniform random value in [-0.05..0.05]
    """
    def __init__(self, *args, **kwargs):
        kwargs["boundary_wall"] = False
        kwargs["ground_color"] = "white"
        super().__init__(*args, **kwargs)
        self.time_step = 0.02
        self._time_decimal_places = 2
        self.gravity = 9.8
        self.cart = Cart(0)
        self.add_robot(self.cart)
        self.cart.reset()



# -*- coding: utf-8 -*-
# ************************************************************
# aitk.robots: Python robot simulator
#
# Copyright (c) 2021 AITK Developers
#
# https://github.com/ArtificialIntelligenceToolkit/aitk.robots
# ************************************************************

import math

from ..utils import (
    display,
    arange,
    distance,
    degrees_to_world,
    rotate_around,
    world_to_degrees,
    ONE80_OVER_PI,
    PI_OVER_180,
    PI_OVER_2,
)
from .base import BaseDevice

from aitk.utils import Color

class RangeSensor(BaseDevice):
    def __init__(
        self, position=(8, 0), a=0, max=20, width=1.0, name="sensor", **kwargs
    ):
        """
        A range sensor calculates both a "distance" and a "reading".
        The "distance" is the actual distance to an obstacle in CM (when
        it is within the "max" distance).
        The reading is a ratio of distance/max. When no obstacle is sensed 
        the "reading" is 1.0 and as the robot approaches an obstacle the 
        "reading" decreases towards 0.0.

        Args:
            position (List[int, int]): the location on the robot in (x, y)
            a (float): the direction in degrees the sensor is
                facing.
            max (float): max distance in CM that the range sensor can sense
            width (float): 0 for laser, or wider for sonar
            name (str): the name of the sensor
        """
        config = {
            "position": position,
            "a": a, # degrees in the config file
            "max": max,
            "width": width,
            "name": name,
            "visible": True,
        }
        config.update(kwargs)
        self._watcher = None
        self.robot = None
        self.initialize()
        self.from_json(config)

    def initialize(self):
        """
        Internal method to set all settings to default values.
        """
        self.type = "ir"
        self.time = 0.0
        self.reading = 1.0
        self.position = [8, 0]
        self.dist_from_center = distance(0, 0, self.position[0], self.position[1])
        self.dir_from_center = math.atan2(-self.position[0], self.position[1])
        self.a = degrees_to_world(0)  # comes in degrees, save as radians
        self.max = 20  # CM
        self.width = 1.0  # radians
        self.visible = True
        self.name = "sensor"
        self.distance = self.reading * self.max

    def watch(self, title="Range Sensor:"):
        """
        Create a dynamically updating view
        of this sensor.

        Args:
            title (str): title of sensor
        """
        widget = self.get_widget(title=title)
        return display(widget)

    def get_widget(self, title=None, attributes=None):
        """
        Return the dynamically updating widget.

        Args:
            title (str): title of sensor
        """
        from ..watchers import AttributesWatcher

        if self.robot is None or self.robot.world is None:
            print("ERROR: can't watch until added to robot, and robot is in world")
            return None

        if self._watcher is None:
            title = title if title is not None else "Range Sensor:"
            attributes = attributes if attributes is not None else "all"
            self._watcher = AttributesWatcher(
                self,
                "name",
                "reading",
                "distance",
                title=title,
                labels=["Name:", "Reading:", "Distance:"],
                attributes=attributes,
            )
            self.robot.world._watchers.append(self._watcher)
        else:
            self._watcher.set_arguments(title=title, attributes=attributes)
            self._watcher.update()
        return self._watcher.widget

    def from_json(self, config):
        """
        Set the settings from a device config.

        Args:
            config (dict): a config dictionary
        """
        valid_keys = set([
            "position", "a", "max", "width", "name", "class", "visible"
        ])
        self.verify_config(valid_keys, config)

        if "position" in config:
            self.position = config["position"]
            # Get location of sensor, doesn't change once position is set:
            self.dist_from_center = distance(0, 0, self.position[0], self.position[1])
            self.dir_from_center = math.atan2(-self.position[0], self.position[1])
        if "a" in config:
            self.a = degrees_to_world(config["a"])
        if "max" in config:
            self.max = config["max"]
        if "width" in config:
            self.width = config["width"] * PI_OVER_180  # save as radians
        if "visible" in config:
            self.visible = config["visible"]
        if self.width == 0:
            self.type = "laser"
        if "name" in config:
            self.name = config["name"]
        self.distance = self.reading * self.max

    def to_json(self):
        """
        Save the internal settings to a config dictionary.
        """
        config = {
            "class": self.__class__.__name__,
            "position": self.position,
            "a": world_to_degrees(self.a),
            "max": self.max,
            "width": self.width * ONE80_OVER_PI,  # save as degrees
            "visible": self.visible,
            "name": self.name,
        }
        return config

    def __repr__(self):
        return "<RangeSensor %r a=%r, max=%r, width=%r, position=%r>" % (
            self.name,
            round(world_to_degrees(self.a), 1),
            self.max,
            round(self.width * ONE80_OVER_PI, 1),
            self.position,
        )

    def _step(self, time_step):
        pass

    def update(self, draw_list=None):
        """
        Update the device.

        Args:
            draw_list (list): optional. If given, then the
                method can add to it for drawing later.
        """
        # Update timestamp:
        self.time = self.robot.world.time
        # This changes:
        p = rotate_around(
            self.robot.x,
            self.robot.y,
            self.dist_from_center,
            self.robot.a + self.dir_from_center + PI_OVER_2,
        )

        if self.robot.world.debug and draw_list is not None:
            draw_list.append(("draw_ellipse", (p[0], p[1], 2, 2), {}))

        self.set_reading(1.0)
        if self.width != 0:
            for incr in arange(-self.width / 2, self.width / 2, self.width / 2):
                hits = self.robot.cast_ray(
                    p[0],
                    p[1],
                    -self.robot.a + PI_OVER_2 + incr - self.a,
                    self.max,
                )
                if hits:
                    if self.robot.world.debug and draw_list is not None:
                        draw_list.append(
                            ("draw_ellipse", (hits[-1].x, hits[-1].y, 2, 2), {})
                        )
                    # Closest hit:
                    if hits[-1].distance < self.get_distance():
                        self.set_distance(hits[-1].distance)
        else:
            hits = self.robot.cast_ray(
                p[0],
                p[1],
                -self.robot.a + PI_OVER_2 - self.a,
                self.max,
            )
            if hits:
                if self.robot.world.debug and draw_list is not None:
                    draw_list.append(("draw_ellipse", (hits[-1].x, hits[-1].y, 2, 2), {}))
                # Closest hit:
                if hits[-1].distance < self.get_distance():
                    self.set_distance(hits[-1].distance)

    def draw(self, backend):
        """
        Draw the device on the backend.

        Args:
            backend (Backend): an aitk drawing backend
        """
        if not self.visible:
            return
        
        backend.set_fill(Color(128, 0, 128, 64))
        dist = self.get_distance()
        if self.width > 0:
            if self.get_reading() < 1.0:
                backend.strokeStyle(Color(255), 1)
            else:
                backend.strokeStyle(Color(0), 1)

            backend.draw_arc(
                self.position[0],
                self.position[1],
                dist,
                dist,
                self.a - self.width / 2,
                self.a + self.width / 2,
            )
        else:
            if self.get_reading() < 1.0:
                backend.strokeStyle(Color(255), 1)
            else:
                backend.strokeStyle(Color(128, 0, 128, 64), 1)

            x, y = rotate_around(self.position[0], self.position[1], dist, self.a)

            backend.draw_line(
                self.position[0], self.position[1], x, y
            )

    def get_distance(self):
        """
        Get the last range distance of the sensor in CM. The distance is
        between 0 and max.
        """
        return self.distance

    def get_reading(self):
        """
        Get the last range reading of the sensor.  The reading is between
        0 and 1.
        """
        return self.reading

    def get_max(self):
        """
        Get the maximum distance in CM the sensor can sense.
        """
        return self.max

    def get_position(self):
        """
        Get the position of the sensor. This represents the location
        of the sensor in [x, y] CM.
        """
        return self.position

    def get_angle(self):
        """
        Get the direction in degrees. Use RangeSensor.a
        to get the raw radians.
        """
        return world_to_degrees(self.a)

    def get_width(self):
        """
        Get the width of the sensor in degrees. Use
        RangeSensor.width to see raw radians.
        """
        return self.width * ONE80_OVER_PI

    def get_name(self):
        """
        Get the name of the range sensor.
        """
        return self.name

    def set_name(self, name):
        """
        Set the name of the range sensor.

        Args:
            name (str): the name of the range sensor
        """
        self.name = name

    def set_distance(self, distance):
        """
        Set the distance that the sensor is reading. You would not
        usually do this manually.

        Args:
            distance (float): distance in CM to sensed object
        """
        self.distance = distance
        self.reading = distance / self.max

    def set_reading(self, reading):
        """
        Set the reading that the sensor is reading. You would not
        usually do this manually.

        Args:
            reading (float): between 0 and 1
        """
        self.reading = reading
        self.distance = reading * self.max

    def set_max(self, max):
        """
        Set the maximum distance in CM that this sensor can sense.

        Args:
            max (float): max distance in CM the sensor can sense.
        """
        self.max = max

    def set_position(self, position):
        """
        Set the position of the sensor. position must be a
        list/tuple of length 2 representing [x, y] in CM of the
        location of the sensor relative to the center of the
        robot.

        Args:
            position (List[int, int]): the location
                of the sensor in relationship to the center of the
                robot.
        """
        if len(position) != 2:
            raise ValueError("position must be of length two")

        self.position = position
        self.dist_from_center = distance(0, 0, self.position[0], self.position[1])
        self.dir_from_center = math.atan2(-self.position[0], self.position[1])

    def set_angle(self, a):
        """
        Set the direction of the sensor.

        Args:
            a (float): the angle of the direction of sensor in degrees
        """
        self.a = degrees_to_world(a)

    def set_width(self, width):
        """
        Set the width of the range sensor in degrees. 0 width
        is a laser range finder. Larger values indicate the
        width of an IR sensor. It is measured in three locations:
        start, middle, and stop. The value of the sensor is the
        minimum of the three.

        Args:
            width (float): angle in degrees
        """
        self.width = width * PI_OVER_180  # save as radians
        if self.width == 0:
            self.type = "laser"
        else:
            self.type = "ir"

    def set_visible(self, visible):
        """
        Set the display visibility of the rangesensor.
        """
        self.visible = visible



# -*- coding: utf-8 -*-
# ******************************************************
# aitk.networks: Keras model wrapper with visualizations
#
# Copyright (c) 2021-2024 Douglas S. Blank
#
# https://github.com/ArtificialIntelligenceToolkit/aitk.networks
#
# ******************************************************

import functools
import html
import io
import itertools
import math
import numbers
import operator
import random
from types import FunctionType

import matplotlib.pyplot as plt
import numpy as np
import tensorflow as tf
import tensorflow.keras.backend as K
from matplotlib import cm
from PIL import Image
from tensorflow.keras.layers import Concatenate, Dense, InputLayer, Layer
from tensorflow.keras.models import Model

from aitk.utils import array_to_image

from .utils import (
    get_argument_bindings,
    get_array_shape,
    get_connections,
    get_error_colormap,
    get_layer_input_tensor,
    get_templates,
    image_to_uri,
    is_keras_tensor,
    make_input_from_shape,
    render_curve,
    svg_to_image,
)

try:
    from IPython.display import HTML, clear_output, display
except ImportError:
    HTML = None


class Network:
    """
    Wrapper around a keras.Model.
    """

    def __init__(self, model=None, layers=None, name="Network", **config):
        if model is not None and layers is not None:
            raise Exception("Network() takes model or layers, not both")

        self._state = {
            "tolerance_accuracy_used": False,
            "pca": {},
        }
        self._watchers = []
        self._fit_inputs = None
        self._fit_targets = None
        self._connections = []
        self._model = None
        # Place to put models between layers:
        self._predict_models = {}
        # Place to map layer to its input layers:
        self._input_layer_names = {}
        # For saving HTML for watchers
        self._svg = None
        self._history = {"weights": [], "metrics": []}
        self._epoch = 0
        self._tolerance = 0.1
        self._name = name
        self._layers = []
        self._layers_map = {}
        self.input_bank_order = []
        self.output_bank_order = []
        self._level_ordering = []
        self.config = {
            "name": self._name,  # for svg title
            "class_id": "keras-network",  # for svg network classid
            "id": "keras-network",  # for svg id
            "font_size": 12,  # for svg
            "font_family": "monospace",  # for svg
            "font_color": "black",  # for svg
            "font_color_dynamic": "red",  # for svg
            "font_color_not_visible": "green",  # for svg
            "font_color_layer_name": "blue",  # for svg
            "border_top": 25,  # for svg
            "border_bottom": 25,  # for svg
            "hspace": 200,  # for svg
            "vspace": 30,  # for svg, arrows
            "max_pixels": 200,  # for svg
            "pixels": 50,  # for svg
            "activation": "linear",  # Dense default, if none specified
            "arrow_color": "black",
            "arrow_width": "2",
            "border_width": "2",
            "border_color": "black",
            "background_color": "#B0C4DE",  # must use hex format here
            "show_targets": False,
            "show_error": False,
            "precision": 2,
            "scale": None,  # for svg, 0 - 1, or None for optimal
            "rotate": False,  # for rotating SVG
            "smoothing": 0.02,  # smoothing curves
            "preferred_size": 400,  # in pixels
            "max_width": 800,  # in pixels
            "layers": {},
            # layer_name: {vshape, feature, keep_aspect_ratio, visible
            # colormap, border_color, border_width}
        }
        # Override settings:
        self.set_config(**config)
        if model:
            self._model = model
            self._name = self._model.name
            for layer in model.layers:
                self.add(layer)
            self._connections = get_connections(model)
            self.compile()
        elif layers:
            for layer in layers:
                self.add(layer)
        # When we are done here, we are in 1 of 2 states:
        # 1. A model, ready to go
        # 2. Network, ready for more add(), connect(), compile()

    def __getattr__(self, attr):
        if self._model is None:
            raise Exception("Model has not yet been compiled")
        return getattr(self._model, attr)

    def __getitem__(self, layer_name):
        return self._layers_map.get(layer_name, None)

    def add(self, layer):
        """
        Add a layer to the network.
        """
        if isinstance(layer, FunctionType):
            raise Exception("Don't use Input; use InputLayer")

        if not isinstance(layer, Layer):
            raise Exception("Network.add() requires a Layer")

        # Let's find a good name for the layer:
        name = layer.name
        if name.startswith("keras_tensor"):
            name = "input" + name[12:]

        if name in self._layers_map:
            raise Exception("The name %r is already used" % name)

        # Add the layer:
        layer.name = name
        self._layers.append(layer)
        self._layers_map[layer.name] = layer

    @property
    def model(self):
        return self._model

    def initialize_model(self):
        # Build intermediary models:
        self._build_predict_models()
        # Config for various layer settings (like 'vshape'):
        self.config["layers"] = {layer.name: {} for layer in self._layers}

    def connect(self, from_layer_name=None, to_layer_name=None):
        """ """
        if len(self._layers) == 0:
            raise Exception("no layers have been added")
        if from_layer_name is not None and not isinstance(from_layer_name, str):
            raise Exception("from_layer_name should be a string or None")
        if to_layer_name is not None and not isinstance(to_layer_name, str):
            raise Exception("to_layer_name should be a string or None")
        if from_layer_name is None and to_layer_name is None:
            for i in range(len(self._layers) - 1):
                from_layer = self._layers[i]
                to_layer = self._layers[i + 1]
                self.connect(from_layer.name, to_layer.name)
        else:
            if from_layer_name == to_layer_name:
                raise Exception("self connections are not allowed")
            if not isinstance(from_layer_name, str):
                raise Exception("from_layer_name should be a string")
            if from_layer_name not in self._layers_map:
                raise Exception("unknown layer: %s" % from_layer_name)
            if not isinstance(to_layer_name, str):
                raise Exception("to_layer_name should be a string")
            if to_layer_name not in self._layers_map:
                raise Exception("unknown layer: %s" % to_layer_name)
            from_layer = self[from_layer_name]
            to_layer = self[to_layer_name]
            # Check for input going to a Dense to warn:
            # if len(from_layer.shape) > 2 and to_layer.__class__.__name__ == "Dense":
            #    print("WARNING: connected multi-dimensional input layer '%s' to layer '%s'; consider adding a FlattenLayer between them" % (
            #        from_layer.name, to_layer.name), file=sys.stderr)
            self._connections.append((from_layer_name, to_layer_name))

    def fit(self, *args, **kwargs):
        """
        Train the model.

        kwargs:
            * accuracy: (float) stop when accuracy gets equal or above this value
            * val_accuracy: (float) stop when validation accuracy gets equal or above this value
            * tolerance: (float) use this value for measuring accuracies
            * report_rate: (int) update the graphs every N epochs
            * loss: (float) stop when loss gets equal or below this value
            * val_loss: (float) stop when validation loss gets equal or below this value
            * patience: (int) number of epochs to wait without improvements until stopping
            * save: (bool) whether to save the weights at end of each epoch
            * monitor: (str) metric to monitor to determine whether to stop
            * callbacks: (list) list of callbacks
        """
        from .callbacks import UpdateCallback, make_early_stop, make_save, make_stop

        # plot = True
        # if plot:
        #    import matplotlib
        #    mpl_backend = matplotlib.get_backend()
        # else:
        #    mpl_backend = None
        # Get any kwargs that are not standard:
        report_rate = kwargs.pop("report_rate", 1)
        # Early stopping and Stop on Accuracy, Val_accuracy
        patience = kwargs.pop("patience", 0)
        # Our stopping criteria:
        accuracy = kwargs.pop("accuracy", None)
        tolerance = kwargs.pop("tolerance", None)
        if tolerance is not None:
            self.set_tolerance(tolerance)
        val_accuracy = kwargs.pop("val_accuracy", None)
        loss = kwargs.pop("loss", None)
        val_loss = kwargs.pop("val_loss", None)
        # Monitoring callbacks:
        # Save weights:
        save = kwargs.pop("save", 0)
        # Keras Early stopping:
        monitor = kwargs.pop("monitor", None)

        update_callback = UpdateCallback(self, report_rate)
        kwargs = get_argument_bindings(self._model.fit, args, kwargs)
        # get callbacks, if any:
        callbacks = kwargs.get("callbacks", None)
        if callbacks is None:
            callbacks = []
        # add our update callback to it:
        callbacks.append(update_callback)
        # add any other callbacks:
        if monitor is not None:
            callbacks.append(make_early_stop(monitor, patience))
        if save > 0:
            callbacks.append(make_save(self, save))
        if accuracy is not None:
            callbacks.append(make_stop("accuracy", accuracy, patience, False))
        if val_accuracy is not None:
            callbacks.append(make_stop("accuracy", val_accuracy, patience, True))
        if loss is not None:
            callbacks.append(make_stop("loss", loss, patience, False))
        if val_loss is not None:
            callbacks.append(make_stop("loss", val_loss, patience, True))
        kwargs["callbacks"] = callbacks
        kwargs["verbose"] = 0
        kwargs["initial_epoch"] = self._epoch

        shape = get_array_shape(kwargs.get("x"))
        # TODO: check all types of networks
        if shape:
            kwargs["x"] = np.array(kwargs["x"])
            kwargs["y"] = np.array(kwargs["y"])

        self._fit_inputs = kwargs.get("x")  # inputs
        self._fit_targets = kwargs.get("y")  # targets

        # call underlying model fit:
        try:
            history = self._model.fit(**kwargs)
        except KeyboardInterrupt:
            plt.close()
            raise KeyboardInterrupt() from None

        if save > 0:
            # FIXME: don't save if didn't go through loop?
            self._history["weights"].append((self._epoch, self.get_weights()))

        metrics = {
            key: history.history[key][-1]
            for key in history.history
            if len(history.history[key]) > 0
        }

        ## FIXME: getting epochs by keyword:
        print(
            "Epoch %d/%d %s"
            % (
                self._epoch,
                kwargs["epochs"],
                " - ".join(
                    ["%s: %s" % (key, value) for (key, value) in metrics.items()]
                ),
            )
        )
        return history

    def in_console(self, mpl_backend: str) -> bool:
        """
        Return True if running connected to a console; False if connected
        to notebook, or other non-console system.

        Possible values:
            * 'TkAgg' - console with Tk
            * 'Qt5Agg' - console with Qt
            * 'MacOSX' - mac console
            * 'module://ipykernel.pylab.backend_inline' - default for notebook and
              non-console, and when using %matplotlib inline
            * 'NbAgg' - notebook, using %matplotlib notebook

        Here, None means not plotting, or just use text.

        Note:
            If you are running ipython without a DISPLAY with the QT
            background, you may wish to:

            export QT_QPA_PLATFORM='offscreen'
        """
        return mpl_backend not in [
            "module://ipykernel.pylab.backend_inline",
            "NbAgg",
        ]

    def on_epoch_end(self, callback, logs, report_rate=None, clear=True):
        """
        plots loss and accuracy on separate graphs, ignoring any other
        metrics for now.

        Args:
            report_rate: if None, then just update the plot, else
                advance self._epoch, and possible update plot
        """
        format = "svg"

        if report_rate is not None:
            # update epoch and history, and then possibly draw
            # The metrics after the epoch of training, so advanced first
            self._epoch += 1
            self._history["metrics"].append((self._epoch, logs))

            if (self._epoch % report_rate) != 0:
                return

        if len(self._watchers) > 0:
            index = random.randint(0, self.get_input_length(self._fit_inputs) - 1)
            inputs = self.get_input_from_dataset(index, self._fit_inputs)
            targets = self.get_target_from_dataset(index, self._fit_targets)
            self.propagate(inputs, targets)  # update watchers

        metrics = [list(history[1].keys()) for history in self._history["metrics"]]
        metrics = set([item for sublist in metrics for item in sublist])

        def match_acc(name):
            return name.endswith("acc") or name.endswith("accuracy")

        def match_val(name):
            return name.startswith("val_")

        if callback._figure is not None:
            # figure and axes objects have already been created
            fig, loss_ax, acc_ax = callback._figure
            loss_ax.clear()
            if acc_ax is not None:
                acc_ax.clear()
        else:
            # first time called, so create figure and axes objects
            if any([match_acc(metric) for metric in metrics]):
                fig, (loss_ax, acc_ax) = plt.subplots(1, 2, figsize=(10, 4))
            else:
                fig, loss_ax = plt.subplots(1)
                acc_ax = None
            callback._figure = fig, loss_ax, acc_ax

        def get_xy(name):
            return [
                (history[0], history[1][name])
                for history in self._history["metrics"]
                if name in history[1]
            ]

        for metric in metrics:
            xys = get_xy(metric)
            x_values = [xy[0] for xy in xys]
            y_values = [xy[1] for xy in xys]
            if metric == "loss":
                loss_ax.plot(x_values, y_values, label=metric, color="r")  # red
            elif metric == "val_loss":
                loss_ax.plot(x_values, y_values, label=metric, color="orange")
            elif match_acc(metric) and not match_val(metric) and acc_ax is not None:
                acc_ax.plot(x_values, y_values, label=metric, color="b")  # blue
            elif match_acc(metric) and match_val(metric) and acc_ax is not None:
                acc_ax.plot(x_values, y_values, label=metric, color="c")  # cyan
            # FIXME: add a chart for each metric
            # else:
            #    loss_ax.plot(x_values, y_values, label=metric)

        loss_ax.set_ylim(bottom=0)
        loss_ax.set_title("%s: Training Loss" % (self.name,))
        loss_ax.set_xlabel("Epoch")
        loss_ax.set_ylabel("Loss")
        loss_ax.legend(loc="best")
        if acc_ax is not None:
            acc_ax.set_ylim([-0.1, 1.1])
            acc_ax.set_title("%s: Training Accuracy" % (self.name,))
            acc_ax.set_xlabel("Epoch")
            acc_ax.set_ylabel("Accuracy")
            acc_ax.legend(loc="best")

        if True or format == "svg":
            # FIXME: work in console
            # if (callback is not None and not callback.in_console) or format == "svg":
            bytes = io.BytesIO()
            plt.savefig(bytes, format="svg")
            img_bytes = bytes.getvalue()
            if HTML is not None:
                if clear:
                    clear_output(wait=True)
                display(HTML(img_bytes.decode()))
            else:
                raise Exception("need to install `IPython` to display matplotlib plots")
        else:  # format is None
            plt.pause(0.01)
            # plt.show(block=False)

    def _prepare_input(self, inputs, input_names):
        """
        Get the input_names from the inputs
        """
        # inputs is either a dict or a list, where index matches
        # the input banks.
        if isinstance(inputs, dict):
            return [np.array(inputs[name]) for name in input_names]
        elif len(self.input_bank_order) == 1:
            return np.array([inputs])
        else:
            return [
                np.array([inputs[index]])
                for index in [self.input_bank_order.index(name) for name in input_names]
            ]

    def build_model(self):
        if len(self._connections) == 0:
            raise Exception("Need to connect layers before building model")

        # Assumes layers either added or passed in via layers
        # and connected via Network.connect()
        froms = [connect[0] for connect in self._connections]
        tos = [connect[1] for connect in self._connections]
        input_layers = []
        output_layers = []
        for layer in self._layers:
            if layer.name not in tos:
                input_layers.append(layer.name)
            if layer.name not in froms:
                output_layers.append(layer.name)
        # Now we build the model:
        outputs = [self._build_graph_to(output_layer) for output_layer in output_layers]
        inputs = [self[layer_name]._input_tensor for layer_name in input_layers]
        self._model = Model(inputs=inputs, outputs=outputs, name=self._name)

    def _get_layers_to(self, layer_name):
        return [
            self[connection[0]]
            for connection in self._connections
            if connection[1] == layer_name
        ]

    def _get_layers_from(self, layer_name):
        return [
            self[connection[1]]
            for connection in self._connections
            if connection[0] == layer_name
        ]

    def topological_sort(self, layers, input_layers):
        for layer in layers:
            layer.visited = False
        # Next gather them:
        sorted_layers = []
        queue = input_layers
        while queue:
            current = queue.pop(0)
            if not current.visited:
                sorted_layers.append(current)
                current.visited = True
                queue.extend(self._get_layers_from(current.name))
        for layer in layers:
            if layer.visited is False:
                raise Exception(
                    "Layer %r is not part of the network graph" % layer.name
                )
        return sorted_layers

    def _build_graph_to(self, layer_name):
        """
        Given the name of a layer, build all of the models
        to that layer by calling the Keras layer as a function.
        """
        # recursive
        layers = self._get_layers_to(layer_name)
        if len(layers) == 0:
            # An input layer:
            return self[layer_name]

        incoming_layers = [
            self._build_graph_to(incoming_layer.name) for incoming_layer in layers
        ]

        if len(incoming_layers) == 1:
            incoming_layer = incoming_layers[0]
        else:  # more than one
            incoming_layer = Concatenate()(
                [get_layer_input_tensor(layer) for layer in incoming_layers]
            )

        if isinstance(incoming_layer, InputLayer):
            incoming_layer = incoming_layer._input_tensor

        layer = self[layer_name]
        return layer(inputs=incoming_layer)

    def compile(self, *args, **kwargs):
        """
        The last step before you run a network.
        """
        # _layers, _connections already set
        self._layers = self.topological_sort(self._layers, self._get_input_layers())
        # Get the input bank names, in order:
        self.input_bank_order = [layer.name for layer in self._get_input_layers()]
        # Get the output bank names, in order:
        self.output_bank_order = [layer.name for layer in self._get_output_layers()]
        # Get the best (shortest path) between layers:
        self._level_ordering = self._get_level_ordering()
        # First, build model if necessary:
        if self._model is None:
            self.build_model()
        # Next, handle any special names:
        metrics = kwargs.pop("metrics", [])
        if len(metrics) > 0:
            metrics = [self.get_metric(metric) for metric in metrics]
            kwargs["metrics"] = metrics
        # Let the standard keras model do the rest:
        results = self._model.compile(*args, **kwargs)
        self.initialize_model()
        return results

    def _post_process_outputs(self, outputs, return_type):
        def numpy(item):
            if hasattr(item, "numpy"):
                return item.numpy()
            else:
                return item

        if len(self.output_bank_order) == 1:
            if return_type == "list":
                return numpy(outputs)[0].tolist()
            elif return_type == "numpy":
                return numpy(outputs)[0]
        else:
            if return_type == "list":
                return [numpy(item)[0].tolist() for item in outputs]
            elif return_type == "numpy":
                return [numpy(item)[0] for item in outputs]

    def _post_process_dataset_outputs(self, outputs, return_type):
        def numpy(item):
            if hasattr(item, "numpy"):
                return item.numpy()
            else:
                return item

        if len(self.output_bank_order) == 1:
            if return_type == "list":
                return numpy(outputs).tolist()
            elif return_type == "numpy":
                return numpy(outputs)
        else:
            if return_type == "list":
                return [numpy(item).tolist() for item in outputs]
            elif return_type == "numpy":
                return [numpy(item) for item in outputs]

    def propagate(self, inputs, return_type="list"):
        """
        Propagate input patterns to a bank in the network.
        """
        if self._model is None:
            raise Exception("Model has not yet been compiled")

        input_vectors = self._prepare_input(inputs, self.input_bank_order)
        try:
            outputs = self._model(input_vectors, training=False)
        except Exception:
            input_layers_shapes = [
                self._get_raw_output_shape(layer_name)
                for layer_name in self.input_bank_order
            ]
            hints = ", ".join(
                [
                    ("%s: %s" % (name, shape))
                    for name, shape in zip(self.input_bank_order, input_layers_shapes)
                ]
            )
            raise Exception(
                "You must supply the inputs for these banks in order and in the right shape: %s"
                % hints
            ) from None

        return self._post_process_outputs(outputs, return_type)

    def set_pca_spaces(self, inputs):
        """
        Make the pca_space with the current weights.
        """
        from sklearn.decomposition import PCA

        for layer in self.layers:
            pca = PCA(2)
            hidden_raw = self.predict_to(inputs, layer.name, return_type="numpy")
            try:
                pca_space = pca.fit(hidden_raw)
            except ValueError:
                # Only one hidden layer unit
                pca_space = None
            self._state["pca"][layer.name] = pca_space

    def get_input_length(self, inputs):
        if len(self.input_bank_order) == 1:
            return len(inputs)
        else:
            return len(inputs[0])

    def predict_histogram_to(self, inputs, layer_name):
        """
        Entire dataset
        """
        if self._model is None:
            raise Exception("Model has not yet been compiled")

        hidden_raw = self.predict_to(inputs, layer_name, return_type="numpy")

        plt.hist(hidden_raw)
        plt.axis("off")
        fp = io.BytesIO()
        plt.savefig(fp, format="png")
        plt.close()
        image = Image.open(fp)
        return image

    def predict_pca_to(self, inputs, layer_name, colors, sizes):
        if self._model is None:
            raise Exception("Model has not yet been compiled")

        if layer_name not in self._state["pca"]:
            raise Exception("Need to set_pca_spaces first")

        hidden_raw = self.predict_to(inputs, layer_name, return_type="numpy")
        pca_space = self._state["pca"][layer_name]
        if pca_space is not None:
            hidden_pca = pca_space.transform(hidden_raw)
            x = hidden_pca[:, 0]
            y = hidden_pca[:, 1]
        else:
            # Only one hidden layer unit; we'll use zeros for Y axis
            x = hidden_raw
            y = np.zeros(len(hidden_raw))

        plt.scatter(x, y, c=colors, s=sizes)
        plt.axis("off")
        fp = io.BytesIO()
        plt.savefig(fp, format="png")
        plt.close()
        image = Image.open(fp)
        return image

    def predict_pca(
        self,
        inputs=None,
        targets=None,
        show_error=False,
        show_targets=False,
        return_type=None,
        rotate=False,
        scale=None,
        clear=True,
        set_pca_spaces=True,
        colors=None,
        sizes=None,
        **config,
    ):
        """ """
        if self._model is None:
            raise Exception("Model has not yet been compiled")

        # This are not sticky; need to set each time:
        config["rotate"] = rotate
        config["scale"] = scale
        # Everything else is sticky:
        self.config.update(config)

        if set_pca_spaces:
            self.set_pca_spaces(inputs)

        try:
            svg = self.to_svg(
                inputs=inputs, targets=targets, mode="pca", colors=colors, sizes=sizes
            )
        except KeyboardInterrupt:
            raise KeyboardInterrupt() from None

        if return_type is None:
            try:
                get_ipython()  # noqa: F821
                return_type = "html"
            except Exception:
                return_type = "image"

        if return_type == "html":
            if HTML is not None:
                if clear:
                    clear_output(wait=True)
                display(HTML(svg))
            else:
                raise Exception(
                    "need to install `IPython` or use Network.predict_pca(return_type='image')"
                )
        elif return_type == "svg":
            return svg
        elif return_type == "image":
            return svg_to_image(svg, self.config)
        else:
            raise ValueError("unable to convert to return_type %r" % return_type)

    def predict_histogram(
        self,
        inputs=None,
        targets=None,
        show_error=False,
        show_targets=False,
        return_type=None,
        rotate=False,
        scale=None,
        clear=True,
        **config,
    ):
        """ """
        if self._model is None:
            raise Exception("Model has not yet been compiled")

        # This are not sticky; need to set each time:
        config["rotate"] = rotate
        config["scale"] = scale
        # Everything else is sticky:
        self.config.update(config)

        try:
            svg = self.to_svg(inputs=inputs, targets=targets, mode="histogram")
        except KeyboardInterrupt:
            raise KeyboardInterrupt() from None

        if return_type is None:
            try:
                get_ipython()  # noqa: F821
                return_type = "html"
            except Exception:
                return_type = "image"

        if return_type == "html":
            if HTML is not None:
                if clear:
                    clear_output(wait=True)
                display(HTML(svg))
            else:
                raise Exception(
                    "need to install `IPython` or use Network.predict_pca(return_type='image')"
                )
        elif return_type == "svg":
            return svg
        elif return_type == "image":
            return svg_to_image(svg, self.config)
        else:
            raise ValueError("unable to convert to return_type %r" % return_type)

    def predict_to(self, inputs, layer_name, return_type="list"):
        """
        Propagate input patterns to a bank in the network.

        Args:
            inputs (dataset): a dataset of inputs
            layer_name (str): layer name to predict to

        Returns:
            a numpy array
        """
        if self._model is None:
            raise Exception("Model has not yet been compiled")

        input_names = self._input_layer_names[layer_name]
        model = self._predict_models[input_names, layer_name]
        inputs = self._prepare_dataset_inputs(inputs)
        try:
            outputs = model(inputs, training=False)
        except Exception:
            input_layers_shapes = [
                self._get_raw_output_shape(layer_name) for layer_name in input_names
            ]
            hints = ", ".join(
                [
                    ("%s: %s" % (name, shape))
                    for name, shape in zip(input_names, input_layers_shapes)
                ]
            )
            raise Exception(
                "You must supply the inputs for these banks in order and in the right shape: %s"
                % hints
            ) from None

        return self._post_process_dataset_outputs(outputs, return_type)

    def predict_from(self, inputs, from_layer_name, to_layer_name):
        """
        Propagate patterns from one bank to another bank in the network.
        """
        if self._model is None:
            raise Exception("Model has not yet been compiled")

        key = (tuple([from_layer_name]), to_layer_name)
        if key not in self._predict_models:
            from_layer = self[from_layer_name]
            path = self.find_path(from_layer, to_layer_name)
            if path is None:
                raise Exception(
                    "no path between %r to %r" % (from_layer_name, to_layer_name)
                )
            # Input should be what next layer expects:
            input_shape = self[path[0]]._build_shapes_dict["input_shape"]
            current = input_layer = make_input_from_shape(input_shape)
            for layer_name in path:
                current = self[layer_name](current)
            self._predict_models[key] = Model(inputs=input_layer, outputs=current)
        model = self._predict_models[key]
        return model(inputs, training=False).numpy()

    def display_colormap(self, colormap=None):
        """
        Display one, or all of the colormaps available on your system.
        """
        if colormap is None:
            if HTML:
                for colorname in cm.cmap_d.keys():
                    display(colorname, self.display_colormap(colorname))
            else:
                raise Exception("you need to install IPython for this function to work")
        else:
            width, height = (400, 25)
            vector = np.arange(0, 1, 0.01)
            vector = vector.reshape((1, 100))
            cm_hot = cm.get_cmap(colormap)
            vector = cm_hot(vector)
            vector = np.uint8(vector * 255)
            image = Image.fromarray(vector)
            image = image.resize((width, height), 0)
            return image

    def get_image(
        self,
        inputs=None,
        targets=None,
        show_error=False,
        show_targets=False,
        return_type="image",
        rotate=False,
        scale=None,
        **config,
    ):
        """
        Create an SVG of the network given some inputs (optional).

        Inputs and targets are for a single pattern.

        Arguments:
            inputs: input values to propagate
            targets: target values to show
            show_error (bool): show the output error in resulting picture
            show_targets (bool): show the targets in resulting picture
            return_type (str): optional "html", "image", or "svg"

        Examples:
            >>> net = Network("Picture", 2, 2, 1)
            >>> net.compile(error="mse", optimizer="adam")
            >>> net.display()
            <IPython.core.display.HTML object>
            >>> net.display([.5, .5])
            <IPython.core.display.HTML object>
            >>> net.display([.5, .5])
            <IPython.core.display.HTML object>
        """
        # This are not sticky; need to set each time:
        config["rotate"] = rotate
        config["scale"] = scale
        # Everything else is sticky:
        self.config.update(config)

        try:
            svg = self.to_svg(
                inputs=inputs,
                targets=targets,
                mode="activation",
                colors=None,
                sizes=None,
            )
        except KeyboardInterrupt:
            raise KeyboardInterrupt() from None

        if return_type == "svg":
            return svg
        elif return_type == "image":
            return svg_to_image(svg, self.config)
        else:
            raise ValueError("unable to convert to return_type %r" % return_type)

    def display(
        self,
        inputs=None,
        targets=None,
        show_error=False,
        show_targets=False,
        return_type=None,
        rotate=False,
        scale=None,
        clear=True,
        **config,
    ):
        if self._model is None:
            raise Exception("Model has not yet been compiled")

        if return_type is None:
            try:
                get_ipython()  # noqa: F821
                return_type = "html"
            except Exception:
                return_type = "image"

        # input_vectors = self._prepare_input(inputs, self.input_bank_order)

        if return_type == "html":
            svg = self.get_image(
                inputs,
                targets,
                show_error,
                show_targets,
                "svg",
                rotate,
                scale,
                **config,
            )
            if HTML is not None:
                if clear:
                    clear_output(wait=True)
                display(HTML(svg))
            else:
                raise Exception(
                    "need to install `IPython` or use Network.display(return_type='image')"
                )
        else:
            image = self.get_image(
                inputs,
                targets,
                show_error,
                show_targets,
                return_type,
                rotate,
                scale,
                **config,
            )
            return image

    def watch_weights(self, to_name):
        """ """
        from .watchers import WeightWatcher

        name = "WeightWatcher: to %s" % (to_name,)
        names = [watcher.name for watcher in self._watchers]
        if name not in names:
            watcher = WeightWatcher(self, to_name)
            self._watchers.append(watcher)
        else:
            watcher = self._watchers[names.index(name)]

        display(watcher._widget)

    def watch_layer(self, layer_name):
        """ """
        from .watchers import LayerWatcher

        name = "LayerWatcher: %s" % (layer_name,)
        names = [watcher.name for watcher in self._watchers]
        if name not in names:
            watcher = LayerWatcher(self, layer_name)
            self._watchers.append(watcher)
        else:
            watcher = self._watchers[names.index(name)]

        display(watcher._widget)

    def watch(
        self,
        show_error=None,
        show_targets=None,
        rotate=None,
        scale=None,
    ):
        """ """
        from .watchers import NetworkWatcher

        name = "NetworkWatcher"
        names = [watcher.name for watcher in self._watchers]
        if name not in names:
            watcher = NetworkWatcher(self, show_error, show_targets, rotate, scale)
            self._watchers.append(watcher)
            widget = watcher._widget
        else:
            watcher = self._watchers[names.index(name)]
            widget = watcher.get_widget(show_error, show_targets, rotate, scale)
        display(widget)

    def predict(
        self,
        inputs,
        targets=None,
        show=True,
    ):
        """
        Update all of the watchers whatever they may be watching,
        and return the propagation of the inputs through the network.
        """
        if show:
            for watcher in self._watchers:
                watcher.update(inputs, targets)
        inputs = self._prepare_dataset_inputs(inputs)
        outputs = self._model(inputs, training=False)
        return outputs

    def propagate_to(
        self,
        inputs,
        layer_name,
        return_type="numpy",
        channel=None,
    ):
        input_names = self._input_layer_names[layer_name]
        model = self._predict_models[input_names, layer_name]
        input_vectors = self._prepare_input(inputs, input_names)
        array = model(input_vectors, training=False)

        if return_type == "image":
            array = self._post_process_outputs(array, "numpy")
            return self._layer_array_to_image(layer_name, array, channel=channel)
        else:
            # Here, return_type is "list" or "numpy"
            return self._post_process_outputs(array, return_type)

    def propagate_each(
        self,
        inputs=None,
        targets=None,
    ):
        """
        Update all of the watchers whatever they may be watching.
        """
        if inputs is not None and targets is not None:
            count = 0
            for ins, targs in self.enumerate_dataset(inputs, targets):
                for watcher in self._watchers:
                    watcher.update(ins, targs)
                    if HTML is not None:
                        clear_output()
                    input("Pattern %s; press enter to continue: " % count)
                    count += 1
        elif inputs is not None:
            count = 0
            for ins in self.enumerate_dataset(inputs):
                for watcher in self._watchers:
                    watcher.update(ins)
                    if HTML is not None:
                        clear_output()
                    input("Pattern %s; press enter to continue: " % count)
                    count += 1

    def _build_predict_models(self):
        # for all layers, inputs to here:
        for layer in self._layers:
            if self._get_layer_type(layer.name) != "input":
                inputs = self._get_input_tensors(layer.name, [])
                input_names = [inp[0] for inp in inputs]
                input_tensors = [inp[1] for inp in inputs]
                self._input_layer_names[layer.name] = tuple(input_names)
                self._predict_models[tuple(input_names), layer.name] = Model(
                    inputs=input_tensors,
                    outputs=self[layer.name].output,  # tensors, tensor
                )
            else:
                self._input_layer_names[layer.name] = tuple([layer.name])
                self._predict_models[tuple([layer.name]), layer.name] = Model(
                    inputs=[layer._input_tensor],
                    outputs=[layer.output],
                )

    def _get_input_tensors(self, layer_name, input_list):
        """
        Given a layer_name, return {input_layer_name: tensor}
        """
        # Recursive; results in input_list of [(name, tensor), ...]
        for layer in self._get_layers_to(layer_name):
            if layer.name in self._input_layer_names:
                for layer_name in self._input_layer_names[layer.name]:
                    if layer_name not in [name for (name, tensor) in input_list]:
                        input_list.append((layer_name, self[layer_name]._input_tensor))
            else:
                if self._get_layer_type(layer.name) == "input":
                    if layer.name not in [name for (name, tensor) in input_list]:
                        input_list.append((layer.name, layer._input_tensor))
                else:
                    self._get_input_tensors(layer.name, input_list)
        return input_list

    def make_image(self, layer_name, vector, colormap=None):
        """
        Given an activation name (or function), and an output vector, display
        make and return an image widget.
        """
        image = self._layer_array_to_image(layer_name, vector)
        # If rotated, and has features, rotate it:
        if self.config.get("rotate", False):
            output_shape = self._get_output_shape(layer_name)
            vshape = self.vshape(layer_name)
            if (isinstance(output_shape, tuple) and len(output_shape) >= 3) or (
                vshape is not None and len(vshape) == 2
            ):
                image = image.rotate(90, expand=1)
        return image

    def _layer_has_channels(self, layer_name):
        class_name = self[layer_name].__class__.__name__
        return class_name in ["Conv2D", "MaxPooling2D"]

    def _layer_array_to_image(self, layer_name, vector, channel=None):
        if self._layer_has_channels(layer_name):
            if channel is None:
                channel = self._get_feature(layer_name)
            select = tuple(
                [slice(None) for i in range(len(vector.shape) - 1)]
                + [slice(channel, channel + 1)]
            )
            vector = vector[select]
        else:
            pass  # let's try it as is

        # If vshape is given, then resize the vector:
        vshape = self.vshape(layer_name)
        if vshape and vshape != vector.shape:
            vector = vector.reshape(vshape)

        try:
            minmax = self._get_dynamic_minmax(layer_name, vector)
            image = array_to_image(vector, minmax=minmax)
        except Exception:
            # Error: make a red image
            image = array_to_image([[[255, 0, 0]], [[255, 0, 0]]])

        return image

    def _get_dynamic_minmax(self, layer_name, vector):
        if self[layer_name].__class__.__name__ == "Dense":
            # Get minmax based on activation function
            minmax = self._get_act_minmax(layer_name)
        elif self[layer_name].__class__.__name__ == "Flatten":
            # Get minmax from previous layer
            inputs_to_layer_name = self._get_layers_to(layer_name)
            minmax = self._get_dynamic_minmax(inputs_to_layer_name[0].name, vector)
        elif self[layer_name].__class__.__name__ == "InputLayer":
            # Hardcoded to typical ranges
            minimum = vector.min()
            maximum = vector.max()
            if minimum < 0:
                minmax = [-1, 1]
            elif maximum > 100 and maximum <= 255:
                # Assuming image
                minmax = [0, 255]
            else:
                minmax = [0, 1]
        else:
            # Compute minmax based on mean +/- std
            avg = vector.mean()
            std = vector.std()
            minimum = vector.min()
            maximum = vector.max()
            minmax = [max(avg - std, minimum), min(avg + std, maximum)]
        return minmax

    def _make_color(self, item):
        if isinstance(item, numbers.Number):
            return (item, item, item)
        else:
            return tuple(item)

    def _get_input_layers(self):
        layers = []
        for layer_from, layer_to in self._connections:
            if layer_from not in layers:
                layers.append(layer_from)
        for layer_from, layer_to in self._connections:
            if layer_to in layers:
                layers.remove(layer_to)
        return [self._layers_map[name] for name in layers]

    def _get_output_layers(self):
        layers = []
        for layer_from, layer_to in self._connections:
            if layer_to not in layers:
                layers.append(layer_to)
        for layer_from, layer_to in self._connections:
            if layer_from in layers:
                layers.remove(layer_from)
        return [self._layers_map[name] for name in layers]

    def vshape(self, layer_name):
        """
        Find the vshape of layer.
        """
        if (
            layer_name in self.config["layers"]
            and "vshape" in self.config["layers"][layer_name]
        ):
            return self.config["layers"][layer_name]["vshape"]
        else:
            return None

    def _get_output_shape(self, layer_name):
        layer = self[layer_name]
        if (layer._build_shapes_dict is not None) and (
            "input_shape" in layer._build_shapes_dict
        ):
            output_shape = layer.compute_output_shape(
                layer._build_shapes_dict["input_shape"]
            )
        elif hasattr(layer, "batch_shape"):
            output_shape = layer.batch_shape
        else:
            output_shape = layer.output.shape

        if isinstance(output_shape, list):
            return output_shape[0][1:]
        else:
            return output_shape[1:]

    def _get_input_shape(self, layer_name):
        layer = self[layer_name]
        input_shape = layer._build_shapes_dict["input_shape"]
        if isinstance(input_shape, list):
            return input_shape[0][1:]
        else:
            return input_shape[1:]

    def _get_raw_output_shape(self, layer_name):
        layer = self[layer_name]
        if (layer._build_shapes_dict is not None) and (
            "input_shape" in layer._build_shapes_dict
        ):
            output_shape = layer.compute_output_shape(
                layer._build_shapes_dict["input_shape"]
            )
        elif hasattr(layer, "batch_shape"):
            output_shape = layer.batch_shape
        else:
            output_shape = layer.output.shape
        return output_shape

    def _get_feature(self, layer_name):
        """
        Which feature plane is selected to show? Defaults to 0
        """
        if (
            layer_name in self.config["layers"]
            and "feature" in self.config["layers"][layer_name]
        ):
            return self.config["layers"][layer_name]["feature"]
        else:
            return 0

    def _get_keep_aspect_ratio(self, layer_name):
        if (
            layer_name in self.config["layers"]
            and "keep_aspect_ratio" in self.config["layers"][layer_name]
        ):
            return self.config["layers"][layer_name]["keep_aspect_ratio"]
        else:
            return False

    def _get_tooltip(self, layer_name):
        """
        String (with newlines) for describing layer."
        """

        def format_range(minmax):
            minv, maxv = minmax
            if minv == -2:
                minv = "-Infinity"
            else:
                minv = "%.1f" % minv
            if maxv == +2:
                maxv = "+Infinity"
            else:
                maxv = "%.1f" % maxv
            return "(%s, %s)" % (minv, maxv)

        layer = self[layer_name]
        activation = self._get_activation_name(layer)
        retval = "Layer: %s %r" % (
            html.escape(self[layer_name].name),
            layer.__class__.__name__,
        )
        if activation:
            retval += "\nAct function: %s" % activation
            retval += "\nAct output range: %s" % (
                format_range(
                    self._get_act_minmax(layer_name),
                )
            )
        retval += "\nActual minmax: %s" % (
            format_range(
                self._layer_minmax(layer_name),
            )
        )
        retval += "\nShape = %s" % (self._get_raw_output_shape(layer_name),)
        return retval

    def _get_visible(self, layer_name):
        if (
            layer_name in self.config["layers"]
            and "visible" in self.config["layers"][layer_name]
        ):
            return self.config["layers"][layer_name]["visible"]
        else:
            return True

    def _get_colormap(self, layer_name):
        if (
            layer_name in self.config["layers"]
            and "colormap" in self.config["layers"][layer_name]
        ):
            return self.config["layers"][layer_name]["colormap"]
        else:
            return ("gray", -2, 2)

    def _get_activation_name(self, layer):
        if hasattr(layer, "activation"):
            return layer.activation.__name__

    def _layer_minmax(self, layer_name):
        if layer_name in self.config["layers"]:
            colormap_minmax = self.config["layers"][layer_name].get("colormap", None)
            if colormap_minmax is not None:
                return colormap_minmax[1:]
        return self._get_act_minmax(layer_name)

    def _get_act_minmax(self, layer_name):
        """
        Get the activation (output) min/max for a layer.

        Note: +/- 2 represents infinity
        """
        layer = self[layer_name]
        activation = self._get_activation_name(layer)
        if activation in ["tanh", "softsign"]:
            return (-1, +1)
        elif activation in ["sigmoid", "softmax", "hard_sigmoid"]:
            return (0, +1)
        elif activation in ["relu", "elu", "softplus"]:
            return (0, +2)
        elif activation in ["selu", "linear"]:
            return (-2, +2)
        else:  # default, or unknown activation function
            return (0, +2)

    def _get_border_color(self, layer_name):
        if (
            layer_name in self.config["layers"]
            and "border_color" in self.config["layers"][layer_name]
        ):
            return self.config["layers"][layer_name]["border_color"]
        else:
            return self.config["border_color"]

    def _get_border_width(self, layer_name):
        if (
            layer_name in self.config["layers"]
            and "border_width" in self.config["layers"][layer_name]
        ):
            return self.config["layers"][layer_name]["border_width"]
        else:
            return self.config["border_width"]

    def _find_spacing(self, row, ordering, max_width):
        """
        Find the spacing for a row number
        """
        return max_width / (len(ordering[row]) + 1)

    def describe_connection_to(self, layer1, layer2):
        """
        Returns a textual description of the weights for the SVG tooltip.
        """
        retval = "Weights from %s to %s" % (layer1.name, layer2.name)
        for klayer in self._layers:
            if klayer.name == layer2.name:
                weights = klayer.get_weights()
                for w in range(len(klayer.weights)):
                    retval += "\n %s has shape %s" % (
                        klayer.weights[w].name,
                        weights[w].shape,
                    )
        return retval

    def _get_input_layers_in_order(self, layer_names):
        """
        Get the input layers in order
        """
        return [
            layer_name
            for layer_name in self.input_bank_order
            if layer_name in layer_names
        ]

    def get_input_from_dataset(self, index, dataset):
        """
        Get input tensor(s) at index from dataset.
        """
        if len(self.input_bank_order) == 1:
            data = dataset[index]
        else:
            data = [bank[index] for bank in dataset]
        return data

    def get_target_from_dataset(self, index, dataset):
        """
        Get target tensor(s) at index from dataset.
        """
        if len(self.output_bank_order) == 1:
            data = dataset[index]
        else:
            data = [bank[index] for bank in dataset]
        return data

    def enumerate_dataset(self, dataset1, dataset2=None):
        """ "
        Takes a dataset and turns it into individual
        sets of one pattern each.
        """
        count = 0
        while True:
            if len(self.input_bank_order) == 1:
                if count < len(dataset1):
                    data1 = dataset1[count]
                else:
                    break
            else:
                if count < len(dataset1[0]):
                    data1 = [bank[count] for bank in dataset1]
                else:
                    break

            if dataset2 is not None:
                if len(self.output_bank_order) == 1:
                    if count < len(dataset2):
                        data2 = dataset2[count]
                    else:
                        break
                else:
                    if count < len(dataset2[0]):
                        data2 = [bank[count] for bank in dataset2]
                    else:
                        break

                yield data1, data2
            else:
                yield data1

            count += 1

    def _prepare_dataset_inputs(self, inputs):
        """
        Take input dataset and make sure it is correct format.
        """
        if len(self.input_bank_order) == 1:
            inputs = np.array(inputs)
        else:
            inputs = [np.array(bank) for bank in inputs]
        return inputs

    def target_to_dataset(self, target):
        # FIXME: handle labels
        if len(self.output_bank_order) == 1:
            targets = [np.array([target])]
        else:
            targets = [np.array([bank]) for bank in target]
        return targets

    def to_svg(
        self, inputs=None, targets=None, mode="activation", colors=None, sizes=None
    ):
        """ """
        # FIXME:
        # First, turn single patterns into a dataset:
        # if inputs is not None:
        #    if mode == "activation":
        #        inputs = self._extract_inputs(inputs, self.input_bank_order)
        #
        # if targets is not None:
        #    if mode == "activation":
        #        # FIXME:
        #        targets = self.target_to_dataset(targets)
        # Next, build the structures:
        struct = self.build_struct(inputs, targets, mode, colors, sizes)
        templates = get_templates(self.config)
        # get the header:
        svg = None
        for template_name, dict in struct:
            if template_name == "head_svg":
                dict["background_color"] = self.config["background_color"]
                svg = templates["head_svg"].format(**dict)
                break
        # build the rest:
        for index in range(len(struct)):
            (template_name, dict) = struct[index]
            # From config:
            dict["class_id"] = self.config["class_id"]
            if template_name != "head_svg" and not template_name.startswith("_"):
                rotate = dict.get("rotate", self.config["rotate"])
                if template_name == "label_svg" and rotate:
                    dict["x"] += 8
                    dict["text_anchor"] = "middle"
                    dict["transform"] = (
                        """ transform="rotate(-90 %s %s) translate(%s)" """
                        % (
                            dict["x"],
                            dict["y"],
                            2,
                        )
                    )
                else:
                    dict["transform"] = ""
                if template_name == "curve":
                    if not dict["drawn"]:
                        curve_svg = render_curve(
                            dict,
                            struct[(index + 1) :],  # noqa: E203
                            templates[template_name],
                            self.config,
                        )
                        svg += curve_svg
                else:
                    t = templates[template_name]
                    svg += t.format(**dict)
        svg += """</svg></g></svg>"""
        return svg

    def build_struct(self, inputs, targets, mode, colors, sizes):
        # inputs, targets are full datasets
        ordering = list(
            reversed(self._level_ordering)
        )  # list of names per level, input to output
        (
            max_width,
            max_height,
            row_heights,
            images,
            image_dims,
        ) = self._pre_process_struct(inputs, ordering, targets, mode, colors, sizes)
        # Now that we know the dimensions:
        struct = []
        cheight = self.config["border_top"]  # top border
        # Display targets?
        if self.config["show_targets"]:
            spacing = self._find_spacing(0, ordering, max_width)
            # draw the row of targets:
            cwidth = 0
            for layer_name, anchor, fname in ordering[0]:  # no anchors in output
                if layer_name + "_targets" not in images:
                    continue
                image = images[layer_name + "_targets"]
                (width, height) = image_dims[layer_name]
                cwidth += spacing - width / 2
                struct.append(
                    [
                        "image_svg",
                        {
                            "name": layer_name + "_targets",
                            "x": cwidth,
                            "y": cheight,
                            "image": image_to_uri(image),
                            "width": width,
                            "height": height,
                            "tooltip": self._get_tooltip(layer_name),
                            "border_color": self._get_border_color(layer_name),
                            "border_width": self._get_border_width(layer_name),
                            "rx": cwidth - 1,  # based on arrow width
                            "ry": cheight - 1,
                            "rh": height + 2,
                            "rw": width + 2,
                        },
                    ]
                )
                # show a label
                struct.append(
                    [
                        "label_svg",
                        {
                            "x": cwidth + width + 5,
                            "y": cheight + height / 2 + 2,
                            "label": "targets",
                            "font_size": self.config["font_size"],
                            "font_color": self.config["font_color"],
                            "font_family": self.config["font_family"],
                            "text_anchor": "start",
                        },
                    ]
                )
                cwidth += width / 2
            # Then we need to add height for output layer again, plus a little bit
            cheight += row_heights[0] + 10  # max height of row, plus some
        # Display error
        if self.config["show_error"]:
            spacing = self._find_spacing(0, ordering, max_width)
            # draw the row of errores:
            cwidth = 0
            for layer_name, anchor, fname in ordering[0]:  # no anchors in output
                if layer_name + "_errors" not in images:
                    continue
                image = images[layer_name + "_errors"]
                (width, height) = image_dims[layer_name]
                cwidth += spacing - (width / 2)
                struct.append(
                    [
                        "image_svg",
                        {
                            "name": layer_name + "_errors",
                            "x": cwidth,
                            "y": cheight,
                            "image": image_to_uri(image),
                            "width": width,
                            "height": height,
                            "tooltip": self._get_tooltip(layer_name),
                            "border_color": self._get_border_color(layer_name),
                            "border_width": self._get_border_width(layer_name),
                            "rx": cwidth - 1,  # based on arrow width
                            "ry": cheight - 1,
                            "rh": height + 2,
                            "rw": width + 2,
                        },
                    ]
                )
                # show a label
                struct.append(
                    [
                        "label_svg",
                        {
                            "x": cwidth + width + 5,
                            "y": cheight + height / 2 + 2,
                            "label": "error",
                            "font_size": self.config["font_size"],
                            "font_color": self.config["font_color"],
                            "font_family": self.config["font_family"],
                            "text_anchor": "start",
                        },
                    ]
                )
                cwidth += width / 2
            # Then we need to add height for output layer again, plus a little bit
            cheight += row_heights[0] + 10  # max height of row, plus some
        # Show a separator that takes no space between output and targets/errors
        if self.config["show_error"] or self.config["show_targets"]:
            spacing = self._find_spacing(0, ordering, max_width)
            # Draw a line for each column in putput:
            cwidth = spacing / 2 + spacing / 2  # border + middle of first column
            # number of columns:
            for level_tups in ordering[0]:
                struct.append(
                    [
                        "line_svg",
                        {
                            "x1": cwidth - spacing / 2,
                            "y1": cheight - 5,  # half the space between them
                            "x2": cwidth + spacing / 2,
                            "y2": cheight - 5,
                            "arrow_color": "green",
                            "tooltip": "",
                        },
                    ]
                )
                cwidth += spacing
        # Now we go through again and build SVG:
        positioning = {}
        level_num = 0
        # For each level:
        hiding = {}
        for row in range(len(ordering)):
            level_tups = ordering[row]
            # how many space at this level for this column?
            spacing = self._find_spacing(row, ordering, max_width)
            cwidth = 0
            # See if there are any connections up:
            any_connections_up = False
            for layer_name, anchor, fname in level_tups:
                if not self._get_visible(layer_name):
                    continue
                elif anchor:
                    continue
                for out in self._get_layers_from(layer_name):
                    if (
                        out.name not in positioning
                    ):  # is it drawn yet? if not, continue,
                        # if yes, we need vertical space for arrows
                        continue
                    any_connections_up = True
            if any_connections_up:
                cheight += self.config["vspace"]  # for arrows
            else:  # give a bit of room:
                # FIXME: determine if there were spaces drawn last layer
                # Right now, just skip any space at all
                # cheight += 5
                pass
            row_height = 0  # for row of images
            # Draw each column:
            for column in range(len(level_tups)):
                (layer_name, anchor, fname) = level_tups[column]
                if not self._get_visible(layer_name):
                    if not hiding.get(
                        column, False
                    ):  # not already hiding, add some space:
                        struct.append(
                            [
                                "label_svg",
                                {
                                    "x": cwidth + spacing - 80,  # center the text
                                    "y": cheight + 15,
                                    "label": "[layer(s) not visible]",
                                    "font_size": self.config["font_size"],
                                    "font_color": self.config["font_color_not_visible"],
                                    "font_family": self.config["font_family"],
                                    "text_anchor": "start",
                                    "rotate": False,
                                },
                            ]
                        )
                        row_height = max(row_height, self.config["vspace"])
                    hiding[column] = True
                    cwidth += spacing  # leave full column width
                    continue
                # end run of hiding
                hiding[column] = False
                # Anchor
                if anchor:
                    anchor_name = "%s-%s-anchor%s" % (layer_name, fname, level_num)
                    cwidth += spacing
                    positioning[anchor_name] = {
                        "x": cwidth,
                        "y": cheight + row_heights[row],
                    }
                    x1 = cwidth
                    # now we are at an anchor. Is the thing that it anchors in the
                    # lower row? level_num is increasing
                    prev = [
                        (oname, oanchor, lfname)
                        for (oname, oanchor, lfname) in ordering[level_num - 1]
                        if (
                            ((layer_name == oname) and (oanchor is False))
                            or (
                                (layer_name == oname)
                                and (oanchor is True)
                                and (fname == lfname)
                            )
                        )
                    ]
                    if prev:
                        tooltip = html.escape(
                            self.describe_connection_to(self[fname], self[layer_name])
                        )
                        if prev[0][1]:  # anchor
                            anchor_name2 = "%s-%s-anchor%s" % (
                                layer_name,
                                fname,
                                level_num - 1,
                            )
                            # draw a line to this anchor point
                            x2 = positioning[anchor_name2]["x"]
                            y2 = positioning[anchor_name2]["y"]
                            struct.append(
                                [
                                    "curve",
                                    {
                                        "endpoint": False,
                                        "drawn": False,
                                        "name": anchor_name2,
                                        "x1": cwidth,
                                        "y1": cheight,
                                        "x2": x2,
                                        "y2": y2,
                                        "arrow_color": self.config["arrow_color"],
                                        "tooltip": tooltip,
                                    },
                                ]
                            )
                            struct.append(
                                [
                                    "curve",
                                    {
                                        "endpoint": False,
                                        "drawn": False,
                                        "name": anchor_name2,
                                        "x1": cwidth,
                                        "y1": cheight + row_heights[row],
                                        "x2": cwidth,
                                        "y2": cheight,
                                        "arrow_color": self.config["arrow_color"],
                                        "tooltip": tooltip,
                                    },
                                ]
                            )
                        else:
                            # draw a line to this bank
                            x2 = (
                                positioning[layer_name]["x"]
                                + positioning[layer_name]["width"] / 2
                            )
                            y2 = (
                                positioning[layer_name]["y"]
                                + positioning[layer_name]["height"]
                            )
                            tooltip = "TODO"
                            struct.append(
                                [
                                    "curve",
                                    {
                                        "endpoint": True,
                                        "drawn": False,
                                        "name": layer_name,
                                        "x1": cwidth,
                                        "y1": cheight,
                                        "x2": x2,
                                        "y2": y2,
                                        "arrow_color": self.config["arrow_color"],
                                        "tooltip": tooltip,
                                    },
                                ]
                            )
                            struct.append(
                                [
                                    "curve",
                                    {
                                        "endpoint": False,
                                        "drawn": False,
                                        "name": layer_name,
                                        "x1": cwidth,
                                        "y1": cheight + row_heights[row],
                                        "x2": cwidth,
                                        "y2": cheight,
                                        "arrow_color": self.config["arrow_color"],
                                        "tooltip": tooltip,
                                    },
                                ]
                            )
                    else:
                        print("that's weird!", layer_name, "is not in", prev)
                    continue
                else:
                    # Bank positioning
                    image = images[layer_name]
                    (width, height) = image_dims[layer_name]
                    cwidth += spacing - (width / 2)
                    positioning[layer_name] = {
                        "name": layer_name
                        + ("-rotated" if self.config["rotate"] else ""),
                        "x": cwidth,
                        "y": cheight,
                        "image": image_to_uri(image),
                        "width": width,
                        "height": height,
                        "tooltip": self._get_tooltip(layer_name),
                        "border_color": self._get_border_color(layer_name),
                        "border_width": self._get_border_width(layer_name),
                        "rx": cwidth - 1,  # based on arrow width
                        "ry": cheight - 1,
                        "rh": height + 2,
                        "rw": width + 2,
                    }
                    x1 = cwidth + width / 2
                y1 = cheight - 1
                # Arrows going up
                for out in self._get_layers_from(layer_name):
                    if out.name not in positioning:
                        continue
                    # draw an arrow between layers:
                    anchor_name = "%s-%s-anchor%s" % (
                        out.name,
                        layer_name,
                        level_num - 1,
                    )
                    # Don't draw this error, if there is an anchor in the next level
                    if anchor_name in positioning:
                        tooltip = html.escape(
                            self.describe_connection_to(self[layer_name], out)
                        )
                        x2 = positioning[anchor_name]["x"]
                        y2 = positioning[anchor_name]["y"]
                        struct.append(
                            [
                                "curve",
                                {
                                    "endpoint": False,
                                    "drawn": False,
                                    "name": anchor_name,
                                    "x1": x1,
                                    "y1": y1,
                                    "x2": x2,
                                    "y2": y2,
                                    "arrow_color": self.config["arrow_color"],
                                    "tooltip": tooltip,
                                },
                            ]
                        )
                        continue
                    else:
                        tooltip = html.escape(
                            self.describe_connection_to(self[layer_name], out)
                        )
                        x2 = (
                            positioning[out.name]["x"]
                            + positioning[out.name]["width"] / 2
                        )
                        y2 = (
                            positioning[out.name]["y"] + positioning[out.name]["height"]
                        )
                        struct.append(
                            [
                                "curve",
                                {
                                    "endpoint": True,
                                    "drawn": False,
                                    "name": out.name,
                                    "x1": x1,
                                    "y1": y1,
                                    "x2": x2,
                                    "y2": y2 + 2,
                                    "arrow_color": self.config["arrow_color"],
                                    "tooltip": tooltip,
                                },
                            ]
                        )
                # Bank images
                struct.append(["image_svg", positioning[layer_name]])
                struct.append(
                    [
                        "label_svg",
                        {
                            "x": positioning[layer_name]["x"]
                            + positioning[layer_name]["width"]
                            + 5,
                            "y": positioning[layer_name]["y"]
                            + positioning[layer_name]["height"] / 2
                            + 2,
                            "label": layer_name,
                            "font_size": self.config["font_size"],
                            "font_color": self.config["font_color_layer_name"],
                            "font_family": self.config["font_family"],
                            "text_anchor": "start",
                        },
                    ]
                )
                output_shape = self._get_output_shape(layer_name)
                if self._layer_has_channels(layer_name):
                    features = str(output_shape[-1])
                    # FIXME:
                    feature = str(self._get_feature(layer_name))
                    if self.config["rotate"]:
                        struct.append(
                            [
                                "label_svg",
                                {
                                    "x": positioning[layer_name]["x"] + 5,
                                    "y": positioning[layer_name]["y"] - 10 - 5,
                                    "label": features,
                                    "font_size": self.config["font_size"],
                                    "font_color": self.config["font_color"],
                                    "font_family": self.config["font_family"],
                                    "text_anchor": "start",
                                },
                            ]
                        )
                        struct.append(
                            [
                                "label_svg",
                                {
                                    "x": positioning[layer_name]["x"]
                                    + positioning[layer_name]["width"]
                                    - 10,
                                    "y": positioning[layer_name]["y"]
                                    + positioning[layer_name]["height"]
                                    + 10
                                    + 5,
                                    "label": feature,
                                    "font_size": self.config["font_size"],
                                    "font_color": self.config["font_color"],
                                    "font_family": self.config["font_family"],
                                    "text_anchor": "start",
                                },
                            ]
                        )
                    else:
                        struct.append(
                            [
                                "label_svg",
                                {
                                    "x": positioning[layer_name]["x"]
                                    + positioning[layer_name]["width"]
                                    + 5,
                                    "y": positioning[layer_name]["y"] + 5,
                                    "label": features,
                                    "font_size": self.config["font_size"],
                                    "font_color": self.config["font_color"],
                                    "font_family": self.config["font_family"],
                                    "text_anchor": "start",
                                },
                            ]
                        )
                        struct.append(
                            [
                                "label_svg",
                                {
                                    "x": positioning[layer_name]["x"]
                                    - (len(feature) * 7)
                                    - 5
                                    - 5,
                                    "y": positioning[layer_name]["y"]
                                    + positioning[layer_name]["height"]
                                    - 5,
                                    "label": feature,
                                    "font_size": self.config["font_size"],
                                    "font_color": self.config["font_color"],
                                    "font_family": self.config["font_family"],
                                    "text_anchor": "start",
                                },
                            ]
                        )
                cwidth += width / 2
                row_height = max(row_height, height)
            cheight += row_height
            level_num += 1
        cheight += self.config["border_bottom"]
        # DONE!
        # Draw the title:
        if mode == "activation":
            title = "Activations for %s" % self.config["name"]
        elif mode == "pca":
            title = "PCAs for %s" % self.config["name"]
        elif mode == "histogram":
            title = "Histograms for %s" % self.config["name"]
        if self.config["rotate"]:
            struct.append(
                [
                    "label_svg",
                    {
                        "x": 10,  # really border_left
                        "y": cheight / 2,
                        "label": title,
                        "font_size": self.config["font_size"] + 3,
                        "font_color": self.config["font_color"],
                        "font_family": self.config["font_family"],
                        "text_anchor": "middle",
                    },
                ]
            )
        else:
            struct.append(
                [
                    "label_svg",
                    {
                        "x": max_width / 2,
                        "y": self.config["border_top"] / 2,
                        "label": title,
                        "font_size": self.config["font_size"] + 3,
                        "font_color": self.config["font_color"],
                        "font_family": self.config["font_family"],
                        "text_anchor": "middle",
                    },
                ]
            )
        # figure out scale optimal, if scale is None
        # the fraction:
        if self.config["scale"] is not None:  # scale is given:
            if self.config["rotate"]:
                scale_value = (self.config["max_width"] / cheight) * self.config[
                    "scale"
                ]
            else:
                scale_value = (self.config["max_width"] / max_width) * self.config[
                    "scale"
                ]
        else:
            if self.config["rotate"]:
                scale_value = self.config["max_width"] / max(cheight, max_width)
            else:
                scale_value = self.config["preferred_size"] / max(cheight, max_width)
        # svg_scale = "%s%%" % int(scale_value * 100)
        scaled_width = max_width * scale_value
        scaled_height = cheight * scale_value
        # Need a top-level width, height because Jupyter peeks at it
        if self.config["rotate"]:
            svg_transform = (
                """ transform="rotate(90) translate(0 -%s)" """ % scaled_height
            )
            # Swap them:
            top_width = scaled_height
            top_height = scaled_width
        else:
            svg_transform = ""
            top_width = scaled_width
            top_height = scaled_height
        struct.append(
            [
                "head_svg",
                {
                    "viewbox_width": int(max_width),  # view port width
                    "viewbox_height": int(cheight),  # view port height
                    "width": int(scaled_width),  # actual pixels of image in page
                    "height": int(scaled_height),  # actual pixels of image in page
                    "id": self.config["id"],
                    "top_width": int(top_width),
                    "top_height": int(top_height),
                    "arrow_color": self.config["arrow_color"],
                    "arrow_width": self.config["arrow_width"],
                    "transform": svg_transform,
                },
            ]
        )
        return struct

    def _get_layer_type(self, layer_name):
        """
        Determines whether a layer is a "input", "hidden", or "output"
        layer based on its connections. If no connections, then it is
        "unconnected".
        """
        incoming_connections = self._get_layers_to(layer_name)
        outgoing_connections = self._get_layers_from(layer_name)
        if len(incoming_connections) == 0 and len(outgoing_connections) == 0:
            return "unconnected"
        elif len(incoming_connections) > 0 and len(outgoing_connections) > 0:
            return "hidden"
        elif len(incoming_connections) > 0:
            return "output"
        else:
            return "input"

    def _get_layer_class(self, layer_name):
        """ """
        layer = self[layer_name]
        return layer.__class__.__name__

    def _get_level_ordering(self):
        """
        Returns a list of lists of tuples from
        input to output of levels.

        Each tuple contains: (layer_name, anchor?, from_name/None)

        If anchor is True, this is just an anchor point.
        """
        # First, get a level for all layers:
        levels = {}
        for layer in self._layers:
            level = max(
                [levels[lay.name] for lay in self._get_layers_to(layer.name)] + [-1]
            )
            levels[layer.name] = level + 1
        max_level = max(levels.values())
        ordering = []
        for i in range(max_level + 1):  # input to output
            layer_names = [
                layer.name for layer in self._layers if levels[layer.name] == i
            ]
            ordering.append(
                [
                    (name, False, [x.name for x in self._get_layers_to(name)])
                    for name in layer_names
                ]
            )  # (going_to/layer_name, anchor, coming_from)
        # promote all output banks to last row:
        for level in range(len(ordering)):  # input to output
            tuples = ordering[level]
            index = 0
            for name, anchor, none in tuples[:]:
                if self._get_layer_type(name) == "output":
                    # move it to last row
                    # find it and remove
                    ordering[-1].append(tuples.pop(index))
                else:
                    index += 1
        # insert anchor points for any in next level
        # that doesn't go to a bank in this level
        # order_cache = {}
        for level in range(len(ordering)):  # input to output
            tuples = ordering[level]
            for name, anchor, fname in tuples:
                if anchor:
                    # is this in next? if not add it
                    next_level = [
                        (n, anchor) for (n, anchor, hfname) in ordering[level + 1]
                    ]
                    if (
                        name,
                        False,
                    ) not in next_level:  # actual layer not in next level
                        ordering[level + 1].append(
                            (name, True, fname)
                        )  # add anchor point
                    else:
                        pass  # finally!
                else:
                    # if next level doesn't contain an outgoing
                    # connection, add it to next level as anchor point
                    for layer in self._get_layers_from(name):
                        next_level = [
                            (n, anchor) for (n, anchor, fname) in ordering[level + 1]
                        ]
                        if (layer.name, False) not in next_level:
                            ordering[level + 1].append(
                                (layer.name, True, name)
                            )  # add anchor point
        ordering = self._optimize_ordering(ordering)
        return ordering

    def _optimize_ordering(self, ordering):
        def perms(items):
            return list(itertools.permutations(items))

        def distance(xy1, xy2):
            return math.sqrt((xy1[0] - xy2[0]) ** 2 + (xy1[1] - xy2[1]) ** 2)

        def find_start(cend, canchor, name, plevel):
            """
            Return position and weight of link from cend/name to
            col in previous level.
            """
            col = 1
            for bank in plevel:
                pend, panchor, pstart_names = bank
                if name == pend:
                    if not panchor and not canchor:
                        weight = 10.0
                    else:
                        weight = 1.0
                    return col, weight
                elif cend == pend and name == pstart_names:
                    return col, 5.0
                col += 1
            raise Exception("connecting layer not found!")

        # First level needs to be in bank_order, and cannot permutate:
        first_level = [(bank_name, False, []) for bank_name in self.input_bank_order]
        perm_count = functools.reduce(
            operator.mul, [math.factorial(len(level)) for level in ordering[1:]]
        )
        if perm_count < 70000:  # globally minimize
            permutations = itertools.product(*[perms(x) for x in ordering[1:]])
            # measure arrow distances for them all and find the shortest:
            best = (10000000, None)
            for ordering in permutations:
                ordering = (first_level,) + ordering
                sum = 0.0
                for level_num in range(1, len(ordering)):
                    level = ordering[level_num]
                    plevel = ordering[level_num - 1]
                    col1 = 1
                    for bank in level:  # starts at level 1
                        cend, canchor, cstart_names = bank
                        if canchor:
                            cstart_names = [cstart_names]  # put in list
                        for name in cstart_names:
                            col2, weight = find_start(cend, canchor, name, plevel)
                            dist = (
                                distance(
                                    (col1 / (len(level) + 1), 0),
                                    (col2 / (len(plevel) + 1), 0.1),
                                )
                                * weight
                            )
                            sum += dist
                        col1 += 1
                if sum < best[0]:
                    best = (sum, ordering)
            return best[1]
        else:  # locally minimize, between layers:
            ordering[0] = first_level  # replace first level with sorted
            for level_num in range(1, len(ordering)):
                best = (10000000, None, None)
                plevel = ordering[level_num - 1]
                for level in itertools.permutations(ordering[level_num]):
                    sum = 0.0
                    col1 = 1
                    for bank in level:  # starts at level 1
                        cend, canchor, cstart_names = bank
                        if canchor:
                            cstart_names = [cstart_names]  # put in list
                        for name in cstart_names:
                            col2, weight = find_start(cend, canchor, name, plevel)
                            dist = (
                                distance(
                                    (col1 / (len(level) + 1), 0),
                                    (col2 / (len(plevel) + 1), 0.1),
                                )
                                * weight
                            )
                            sum += dist
                        col1 += 1
                    if sum < best[0]:
                        best = (sum, level)
                ordering[level_num] = best[1]
            return ordering

    def _pre_process_struct(self, inputs, ordering, targets, mode, colors, sizes):
        """
        Determine sizes and pre-compute images.
        """
        # find max_width, image_dims, and row_height
        # Go through and build images, compute max_width:
        # inputs is full dataset
        row_heights = []
        max_width = 0
        max_height = 0
        images = {}
        image_dims = {}
        # if targets, then need to propagate for error:
        if inputs is None:
            inputs = self.make_dummy_dataset()
        if targets is not None:
            outputs = self.propagate(inputs)
            if len(self.output_bank_order) == 1:
                targets = [targets]
                errors = (np.array(outputs) - np.array(targets)).tolist()
            else:
                errors = []
                for bank in range(len(self.output_bank_order)):
                    errors.append(
                        (np.array(outputs[bank]) - np.array(targets[bank])).tolist()
                    )
        # For each level:
        hiding = {}
        for level_tups in ordering:  # output to input:
            # first make all images at this level
            row_width = 0  # for this row
            row_height = 0  # for this row
            # For each column:
            for column in range(len(level_tups)):
                (layer_name, anchor, fname) = level_tups[column]
                if not self._get_visible(layer_name):
                    if not hiding.get(column, False):
                        row_height = max(
                            row_height, self.config["vspace"]
                        )  # space for hidden indicator
                    hiding[column] = True  # in the middle of hiding some layers
                    row_width += self.config["hspace"]  # space between
                    max_width = max(max_width, row_width)  # of all rows
                    continue
                elif anchor:
                    # No need to handle anchors here
                    # as they occupy no vertical space
                    hiding[column] = False
                    # give it some hspace for this column
                    # in case there is nothing else in this column
                    row_width += self.config["hspace"]
                    max_width = max(max_width, row_width)
                    continue
                hiding[column] = False
                # The rest of this for loop is handling image of bank
                if mode == "pca":
                    image = self.predict_pca_to(inputs, layer_name, colors, sizes)
                elif mode == "histogram":
                    image = self.predict_histogram_to(inputs, layer_name)
                else:  # activations of a dataset
                    try:
                        outputs = self.propagate_to(inputs, layer_name)
                        image = self.make_image(layer_name, outputs)
                    except Exception:
                        # Error: make a red image
                        image = array_to_image(
                            [
                                [
                                    [255, 0, 0],
                                    [255, 0, 0],
                                ]
                            ]
                        )
                (width, height) = image.size
                images[layer_name] = image  # little image
                if self._get_layer_type(layer_name) == "output":
                    if targets is not None:
                        # Target image, targets set above:
                        target_colormap = (
                            "grey",
                            -2,
                            2,
                        )  # FIXME: self[layer_name].colormap
                        target_bank = targets[self.output_bank_order.index(layer_name)]
                        target_array = np.array(target_bank)
                        target_image = self.make_image(
                            layer_name, target_array, target_colormap
                        )
                        # Error image, error set above:
                        error_colormap = (get_error_colormap(), -2, 2)  # FIXME
                        error_bank = errors[self.output_bank_order.index(layer_name)]
                        error_array = np.array(error_bank)
                        error_image = self.make_image(
                            layer_name, error_array, error_colormap
                        )
                        images[layer_name + "_errors"] = error_image
                        images[layer_name + "_targets"] = target_image
                    else:
                        images[layer_name + "_errors"] = image
                        images[layer_name + "_targets"] = image
                max_pixels = self.config["max_pixels"]
                pixels = self.config["pixels"]
                # Keep it within range:
                width = width * pixels
                height = height * pixels
                if width > max_pixels:
                    # scale both down
                    scale = max_pixels / width
                    width = width * scale
                    height = height * scale
                elif height > max_pixels:
                    scale = max_pixels / height
                    height = height * scale
                    width = width * scale
                # Now, make sure either isn't too big or too small:
                width = max(min(max_pixels, width), pixels)
                height = max(min(max_pixels, height), pixels)
                # Record the width and height:
                image_dims[layer_name] = (width, height)
                row_width += width + self.config["hspace"]  # space between
                row_height = max(row_height, height)
            row_heights.append(row_height)
            max_width = max(max_width, row_width)  # of all rows
        return max_width, max_height, row_heights, images, image_dims

    def make_dummy_dataset(self):
        """
        Make a stand-in dataset for this network:
        """
        inputs = []
        for layer_name in self.input_bank_order:
            shape = self._get_input_shape(layer_name)
            if (shape is None) or (isinstance(shape, (list, tuple)) and None in shape):
                v = np.random.rand(100)
            else:
                v = np.random.rand(*shape)
            inputs.append(np.array([v]))
        return inputs

    def set_config(self, **items):
        """
        Set one or more configurable item:
        """
        for item in items:
            if item in self.config:
                if item == "layers":
                    self.set_config_layers(**self.config["layers"])
                else:
                    self.config[item] = items[item]
            else:
                raise AttributeError("no such config item: %r" % item)

    def set_config_layers_by_class(self, class_name, **items):
        """
        Set one or more configurable items in layers
        identified by layer instance type.

        Examples:
        ```python
        >>> net.set_config_layers_by_class("Conv", colormap=("RdGy", -1, 1))
        ```
        """
        for layer in self._layers:
            if layer.__class__.__name__.lower().startswith(class_name.lower()):
                self.set_config_layer(layer.name, **items)

    def set_config_layers_by_name(self, name, **items):
        """
        Set one or more configurable items in layers
        identified by layer instance name.

        Examples:
        ```python
        >>> net.set_config_layers_by_name("input", colormap=("RdGy", -1, 1))
        ```
        """
        for layer in self._layers:
            if layer.name.lower().startswith(name.lower()):
                self.set_config_layer(layer.name, **items)

    def set_config_layers(self, **items):
        """
        Set one or more configurable items in all layers:

        >>> net.set_config_layers(colormap=("plasma", -1, 0))

        """
        for layer in self._layers:
            self.set_config_layer(layer.name, **items)

    def set_config_layer(self, layer_name, **items):
        """
        Set one or more configurable items in a layer:
        """
        proper_items = {
            "vshape": ("integer", "integer"),
            "feature": "integer",
            "keep_aspect_ratio": "boolean",
            "visible": "boolean",
            "colormap": ("string", "number", "number"),
            "border_color": "string",
            "border_width": "integer",
        }

        def validate_type(value, format):
            if format == "integer":
                return isinstance(value, int)
            elif format == "number":
                return isinstance(value, (int, float))
            elif format == "string":
                return isinstance(value, str)
            elif format == "boolean":
                return isinstance(value, bool)
            else:
                return all([validate_type(v, f) for v, f in zip(value, format)])

        if layer_name in self.config["layers"]:
            for item in items:
                if item in proper_items:
                    if validate_type(items[item], proper_items[item]):
                        self.config["layers"][layer_name][item] = items[item]
                    else:
                        raise AttributeError(
                            "invalid form for: %r; should be: %s"
                            % (item, proper_items[item])
                        )
                else:
                    raise AttributeError("no such config layer item: %r" % item)
        else:
            raise AttributeError("no such config layer: %r" % layer_name)

    def get_weights_from_history(self):
        """
        Get the weights saved in history.
        """
        return self._history["weights"]

    def get_weights(self, flat=False):
        """
        Return the weights of the entire network.

        Args:
            flat: (bool) if True, return the weights
                as a single dimensional array.
        """
        if flat:
            array = []
            for layer in self._model.layers:
                for weight in layer.get_weights():
                    array.extend(weight.flatten())
            return array
        else:
            return self._model.get_weights()

    def set_weights(self, weights):
        """
        Set the weights in a network.

        Args:
            weights: a list of pairs of weights and biases for each layer,
                or a single (flat) array of values
        """
        if len(weights) > 0 and isinstance(weights[0], numbers.Number):
            current = 0
            for layer in self._model.layers:
                orig = layer.get_weights()
                new_weights = []
                for item in orig:
                    total = functools.reduce(operator.mul, item.shape, 1)
                    w = np.array(weights[current : current + total])
                    new_weights.append(w.reshape(item.shape))
                    current += total
                layer.set_weights(new_weights)
        else:
            self._model.set_weights(weights)

    def set_learning_rate(self, learning_rate):
        """
        Sometimes called `epsilon`.
        """
        if hasattr(self._model, "optimizer") and hasattr(
            self._model.optimizer, "learning_rate"
        ):
            self._model.optimizer.learning_rate = float(learning_rate)
        else:
            print("WARNING: you need to use an optimizer with learning_rate")

    def get_learning_rate(self):
        """
        Sometimes called `epsilon`.
        """
        if hasattr(self._model, "optimizer") and hasattr(
            self._model.optimizer, "learning_rate"
        ):
            return float(self._model.optimizer.learning_rate.value)
        else:
            print("WARNING: you need to use an optimizer with lr")

    def get_metric(self, name):
        if name == "tolerance_accuracy":
            self._state["tolerance_accuracy_used"] = True

            def tolerance_accuracy(targets, outputs):
                return K.mean(
                    K.all(
                        K.less_equal(
                            K.abs(
                                tf.cast(targets, tf.float32)
                                - tf.cast(outputs, tf.float32)
                            ),
                            self._tolerance,
                        ),
                        axis=-1,
                    ),
                    axis=-1,
                )

            return tolerance_accuracy
        else:
            return name

    def get_momentum(self):
        """ """
        if hasattr(self._model, "optimizer") and hasattr(
            self._model.optimizer, "momentum"
        ):
            return float(self._model.optimizer.momentum.numpy())
        else:
            print("WARNING: you need to use an optimizer with momentum")

    def set_momentum(self, momentum):
        """ """
        if hasattr(self._model, "optimizer") and hasattr(
            self._model.optimizer, "momentum"
        ):
            self._model.optimizer.momentum = float(momentum)
        else:
            print("WARNING: you need to use an optimizer with momentum")

    def get_tolerance(self):
        """ """
        if not self._state["tolerance_accuracy_used"]:
            print(
                "WARNING: you need Network.compile(metrics=['tolerance_accuracy']) to use tolerance"
            )
        return self._tolerance

    def set_tolerance(self, tolerance):
        """ """
        if not self._state["tolerance_accuracy_used"]:
            print(
                "WARNING: you need Network.compile(metrics=['tolerance_accuracy']) to use tolerance"
            )
        self._tolerance = tolerance

    def find_path(self, from_layer, to_layer_name):
        """
        Breadth-first search to find shortest path
        from from_layer to to_layer_name.

        Returns None if there is no path.
        """
        # No need to put from_layer.name in path:
        path = {}
        path[from_layer.name] = []
        queue = [from_layer]
        while queue:
            current = queue.pop()
            if current.name == to_layer_name:
                return path[current.name]
            else:
                # expand:
                for layer in self._get_layers_from(current.name):
                    path[layer.name] = path[current.name] + [layer.name]
                    queue.append(layer)
        return None


class SimpleNetwork(Network):
    def __init__(
        self,
        *layers,
        name="SequentialNetwork",
        activation="sigmoid",
        loss="mse",
        optimizer="sgd",
        metrics=None,
    ):
        """
        Create a simple, stochasitic gradient descent network of Dense
        layers.

        Args:
            layers (list): a list of layer forms (see below)
            name (str): a name for the network
            activation (str): the name of a default activation function
                if one is not given with layer.
            loss (str): the name of a metric to use to compute loss (eg, error)
            optimizer (str or Optimizer): the name of an optimizer or an actual
                 Optimizer instance
            metrics (list): a list of metrics to compute in addition to
                "accuracy" and "tolerance_accuracy".

        Layers may be given in the following formats:
            * int: the number of units in the layer
            * (int, str): number of units, and activation function (non-input layers
                only)
            * (int, int, ...): (input layers only) the shape of the input patterns
            * keras layer instance: an instance of a keras layer, like Flatten()
        """

        def make_name(index, total):
            if index == 0:
                return "input"
            elif index == total - 1:
                return "output"
            elif index == 1 and total == 3:
                return "hidden"
            else:
                return "hidden_%d" % index

        def make_layer(index, layers, activation):
            if isinstance(layers[index], Layer) or is_keras_tensor(layers[index]):
                return layers[index]
            elif isinstance(layers[index], str) and hasattr(
                tf.keras.layers, layers[index]
            ):
                layer_class = getattr(tf.keras.layers, layers[index])
                return layer_class()
            else:
                name = make_name(index, len(layers))
                if index == 0:
                    size = layers[index]
                    if not isinstance(size, (list, tuple)):
                        size = tuple([size])
                    return InputLayer(size, name=name)
                else:
                    size = layers[index]
                    if isinstance(size, int):
                        activation_function = activation
                    elif len(size) == 2 and isinstance(size[1], str):
                        size, activation_function = size
                    else:
                        raise Exception(
                            "Invalid SquentialNetwork layer representation: %r" % size
                        )
                    return Dense(size, activation=activation_function, name=name)

        layers = [make_layer(index, layers, activation) for index in range(len(layers))]
        super().__init__(layers=layers, name=name)
        for i in range(len(layers) - 1):
            self.connect(layers[i].name, layers[i + 1].name)
        if metrics is None:
            metrics = ["tolerance_accuracy"]
        metrics = [self.get_metric(name) for name in metrics]
        self.compile(
            optimizer=self._make_optimizer(optimizer), loss=loss, metrics=metrics
        )

    def _make_optimizer(self, optimizer):
        # Get optimizer with some defaults
        if optimizer == "sgd":
            return tf.keras.optimizers.SGD(
                learning_rate=0.1, momentum=0.9, nesterov=False, name="SGD"
            )
        else:
            return optimizer

    def clear_watchers(self):
        """
        Remove all of the watchers of this network, including
        weights, etc.
        """
        self._watchers[:] = []


SequentialNetwork = SimpleNetwork
